import yaml

from cloudmesh.common.console import Console
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand, writefile


class Networkdata:
    """
    A builder for content in the network-config file with cloud-init
    https://refactoring.guru/design-patterns/builder

    A Builder class follows the Builder design pattern. This design pattern allows
    users of the class to construct a complex data structure using purely instance
    methods. This is very useful in that the user of the class need not concern
    themselves with the underlying implementation of the class. Consequentially,
    changes to this class can be made without needing a refactor of all existing calls.

    More details on the builder design pattern found here.
    https://refactoring.guru/design-patterns/builder

    Example:
    d = Networkdata()\
    .with_ip(ip="10.1.1.10")\
    .with_gateway(gateway="10.1.1.1")\
    .with_nameservers(nameservers=['8.8.8.8', '8.8.4.4'])\
    .with_defaults()\
    .with_dhcp4(interfaces='wifis', interface='wlan0', dhcp4=True)\
    .with_optional(interfaces='wifis', interface='wlan0', optional=True)\
    .with_access_points(interfaces='wifis', interface='wlan0', ssid='MYSSID',
                        password='MYPASSWORD')

    print(d)

    To write to file:
    d = Networkdata()\
        .with_ip(ip='10.1.1.10')\
        .with_gateway(gateway='10.1.1.1')\
        .with_nameservers(nameservers=['8.8.8.8', '8.8.4.4'])\
        .with_defaults()\
        .with_dhcp4(interfaces='wifis', interface='wlan0', dhcp4=True)\
        .with_optional(interfaces='wifis', interface='wlan0', optional=True)\
        .with_access_points(ssid='MYSSID', password='MYPASSWORD')\
        .write(filename='test.tmp')
    """

    def __init__(self, version=2, default=False):
        # Dict will be dumped into YAML string
        self.content = {"version": 2, "ethernets": {}, "wifis": {}}
        if default:
            self.__default__()

    def __str__(self):
        return yaml.dump(self.content)

    def write(self, filename=None):
        """
        Writes a file to a location. Safe write for files on mounted partitions
        """
        if filename is None:
            raise Exception('filename arg supplied is None')
        tmp_location = path_expand('~/.cloudmesh/network-data.tmp')
        writefile(tmp_location, str(self))
        Console.info(f'Writing to {filename}')
        Shell.run(f'cat {tmp_location} | sudo tee {filename}')

    def with_ip(self, interfaces='ethernets', interface='eth0', ip=None):
        if ip is None:
            raise Exception("ip argument supplied is None")

        # If subnet not specified, default to 255.255.255.0
        if "/" not in ip:
            ip += "/24"

        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['addresses'] = [ip]  # Expects a list value
        self.content[interfaces][interface]['dhcp4'] = 'no'
        return self

    def with_gateway(self, interfaces='ethernets', interface='eth0', gateway=None):
        if gateway is None:
            raise Exception("gateway argument supplied is None")

        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['gateway4'] = gateway
        return self

    def with_nameservers(self, interfaces='ethernets', interface='eth0', nameservers=None):
        if nameservers is None:
            raise Exception("nameservers argument suppliled is None")
        if type(nameservers) != list:
            raise TypeError("Expected type of nameservers to be a list")

        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['nameservers'] = {'addresses': nameservers}
        return self

    def with_dhcp4(self, interfaces='ethernets', interface='eth0', dhcp4=True):
        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['dhcp4'] = dhcp4

        return self

    def with_optional(self, interfaces='ethernets', interface='eth0', optional=True):
        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['optional'] = optional

        return self

    def with_access_points(self, interfaces='wifis', interface='wlan0',
                           ssid=None, password=None):
        if ssid is None:
            raise Exception("ssid argument suppliled is None")
        if password is None:
            raise Exception("password argument suppliled is None")

        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        if 'access-points' in self.content[interfaces][interface]:
            access_points = self.content[interfaces][interface]['access-points']
            access_points[ssid] = {'password': password}
        else:
            self.content[interfaces][interface]['access-points'] = {
                ssid: {'password': password}
            }

        return self

    def with_defaults(self, interfaces='ethernets', interface='eth0'):
        """
        Unsure if this is needed, however these params were included in the default config, so we keep
        """
        if interface not in self.content[interfaces]:
            self.content[interfaces][interface] = {}

        self.content[interfaces][interface]['match'] = {"driver": "bcmgenet smsc95xx lan78xx"}
        self.content[interfaces][interface]['set-name'] = interface
        return self

    def __default__(self):
        """
        Set the default configuration the one that comes burnt with the ubuntu server OS

        Captured with
        $ grep - Fv \\  # /{mountpoint}/network-config
        (Removes comments)
        """
        self.content['ethernets'] = {
            "eth0": {
                "match": {
                    "driver": "bcmgenet smsc95xx lan78xx"
                },
                "set-name": "eth0",
                "dhcp4": True,
                "optional": True
            }
        }



import inspect
import os.path
import time

import PySimpleGUI as sg  # noqa
import oyaml as yaml
from cloudmesh.burn.usb import USB
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.common.Host import Host
from cloudmesh.common.Shell import Shell
from cloudmesh.common.Tabulate import Printer
# import PySimpleGUIWx as sg
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.sudo import Sudo
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand
from cloudmesh.diagram.diagram import Diagram
from cloudmesh.burn.wifi.ssid import get_ssid
from cloudmesh.burn.command.burn import _build_default_inventory


def _execute(command):
    # print(".", end="", flush=True)
    os.system(f"{command} >> burn-gui.log")


def image(name):
    with open(path_expand(name), 'rb') as file:
        return file.read()


window_size = (800, 800)
log_size = (600, 600)
status_width = (10, 1)
name_width = (10, 1)
tag_width = (15, 1)
entry_width = (10, 1)
security_width = (20, 1)

image_tags = {
    "os_raspberryos": {
        "name": "Raspberry OS",
        "manager": "latest-full",
        "worker": "latest-lite",
    },
    "os_ubuntu_64bit_20_04": {
        "name": "Ubuntu 64-bit 20.04",
        "manager": "Ubuntu 64 04",
        "worker": "Ubuntu 64 04"
    },
    "os_ubuntu_64bit_20_10": {
        "name": "Ubuntu 64-bit 20.10",
        "manager": "Ubuntu 64 10",
        "worker": "Ubuntu 64 10"
    }
}


class Gui:

    def __init__(self, hostname=None, ip=None, dryrun=False, no_diagram=False):

        self.dryrun = dryrun or False
        self.hostnames_str = hostname
        self.ips_str = ip
        self.hostnames = hostnames = hostname or "red,red[01-02]"
        self.ips = ips = ip or "10.0.0.[1-3]"
        self.ssid = get_ssid()
        self.imaged = ""
        self.wifipassword = ""
        self.no_diagram = no_diagram

        hostnames = Parameter.expand(hostnames)
        manager, workers = Host.get_hostnames(hostnames)

        if workers is None:
            n = 1
        else:
            n = len(workers) + 1
        if ip is None:
            ips = Parameter.expand(f"10.1.1.[1-{n}]")
        else:
            ips = Parameter.expand(ips)

        # cluster_hosts = tuple(zip(ips, hostnames))

        self.key = path_expand("~/.ssh/id_rsa.pub")

        banner("Parameters", figlet=True)

        print("Manager:      ", manager)
        print("Workers:      ", workers)
        print("IPS:          ", ips)
        print("Key:          ", self.key)
        print("Dryrun:       ", self.dryrun)

        self.manager = manager
        self.workers = workers
        self.ips = ips

        self.load_data()

        if not self.no_diagram:
            self.create_diag(self.manager)
        self.create_layout()
        # sg.change_look_and_feel('SystemDefault')
        self.window = sg.Window('Cloudmesh Pi Burn', self.layout, resizable=True, size=window_size)

    def burn(self, kind, hostname):
        """
        if hostname == 'red':
            ip = self.ips[0]
            print("in red")
            print(ip)
        else:
            ip_location = int(hostname.replace('red', ""))
            print("in red00x")
            print(ip_location)
            ip = self.ips[ip_location]
            print("in red")
            print(ip)
        """

        return

    def load_data(self):

        self.background = '#64778d'

        if os_is_mac():
            self.details = USB.get_from_diskutil()
        else:
            self.details = USB.get_from_dmesg()

        self.devices = yaml.safe_load(Printer.write(self.details,
                                                    order=[
                                                        "dev",
                                                        "info",
                                                        "formatted",
                                                        "size",
                                                        "active",
                                                        "readable",
                                                        "empty",
                                                        "direct-access",
                                                        "removable",
                                                        "writeable"],
                                                    header=[
                                                        "Path",
                                                        "Info",
                                                        "Formatted",
                                                        "Size",
                                                        "Plugged-in",
                                                        "Readable",
                                                        "Empty",
                                                        "Access",
                                                        "Removable",
                                                        "Writeable"],
                                                    output="yaml"))

    def create_layout(self):

        def line(msg):
            title = f'__ {msg} '
            title = title.ljust(120, '_')

            burn_layout.append([sg.Text(title)])

        location = os.path.dirname(os.path.abspath(inspect.getsourcefile(Gui)))

        cm_logo = f'{location}/images/cm-logo-100.png'
        pi_logo = f'{location}/images/raspberry-logo-white-100.png'
        rack_file = f"~/.cloudmesh/gui/{self.manager}-rack.png"
        net_file = f"~/.cloudmesh/gui/{self.manager}-net.png"

        burn_layout = [
            [sg.T('')]
        ]

        if not self.no_diagram:
            net_layout = [
                [sg.Image(data=image(net_file), key='net-image', background_color='white')]

            ]
            rack_layout = [
                [sg.Image(data=image(rack_file), key='rack-image', background_color='white')]
            ]
        else:
            net_layout = []
            rack_layout = []

        size = log_size
        log_layout = [[sg.T('', key="log", size=size)]]
        log_scroll_layout = [
            [sg.Column(layout=log_layout, scrollable=True)]
        ]

        burn_layout.append(
            [
                sg.Text(40 * " "),
                sg.Image(data=image(cm_logo), key='cm-logo'),
                sg.Image(data=image(pi_logo), key='pi-logo')
            ]
        )

        #
        # DEVICES
        #
        line("Devices")
        if os_is_mac():
            devices = USB.get_dev_from_diskutil()
        elif os_is_linux():
            devices = USB.get_from_dmesg()
        else:
            devices = {}

        count = 0
        for device in devices:
            default = count == 0
            if os_is_linux():
                burn_layout.append(
                    [sg.Radio(device, group_id="DEVICE",
                              default=default,
                              key=f"device-{device['name']}")]
                )
            else:
                burn_layout.append(
                    [sg.Radio(device, group_id="DEVICE", default=default, key=f"device-{device}")]
                )
            count = count + 1

        #
        # SDCARD
        #
        line("SD Card")

        burn_layout.append(
            [sg.Checkbox("format", default=True, key="imaged")]
        )

        #
        # OPERATING SYSTEM
        #
        line("Operating System")

        count = 0
        for entry in image_tags:
            data = image_tags[entry]
            name = data["name"]
            default = count == 0
            burn_layout.append(
                [sg.Radio(name,
                          group_id="OS",
                          enable_events=True,
                          default=default,
                          key=entry)]
            )
            count = count + 1

        #
        # SECURITY
        #
        line("Security")
        if self.key is not None:
            burn_layout.append([sg.Text("Key", size=security_width), sg.Input(key="key", default_text=self.key)])

        burn_layout.append([sg.Text("SSID", size=security_width), sg.Input(key="ssid", default_text=self.ssid)])
        burn_layout.append(
            [sg.Text("Wifi Password", size=security_width), sg.Input(key="wifi", default_text="", password_char='*')])

        #
        # MANAGER
        #
        line("Manager")

        if self.manager is not None:
            manager = self.manager
            i = 0
            burn_layout.append(
                [
                    sg.Text('', size=status_width, key=str(f'status-{manager}')),
                    sg.Button('Burn', key=str(f'button-{manager}')),
                    sg.Text(manager, size=name_width),
                    sg.Text("manager", size=name_width),
                    sg.Input(default_text=manager, size=name_width, key=str(f'name-{manager}')),
                    sg.Input(default_text=self.ips[i], size=name_width, key=str(f'ip-{manager}')),
                    sg.Text('Image'),
                    sg.Input(default_text="latest-full", size=tag_width, key=str(f'tags-{manager}'))
                ]
            )

        #
        # WORKERS
        #
        line("Workers")

        worker_layout = []
        if self.workers is not None:
            i = 1
            for worker in self.workers:
                worker_layout.append(
                    [
                        sg.Text('', size=status_width, key=str(f'status-{worker}')),
                        sg.Button('Burn', key=str(f'button-{worker}')),
                        sg.Text(worker, size=name_width),
                        sg.Text("worker", size=name_width),
                        sg.Input(default_text=worker, size=name_width, key=str(f'name-{worker}')),
                        sg.Input(default_text=self.ips[i], size=name_width, key=str(f'ip-{worker}')),
                        sg.Text('Image'),
                        sg.Input(default_text="latest-lite", size=tag_width, key=str(f'tags-{worker}')),
                    ])
                i = i + 1
        burn_layout.append([sg.Column(worker_layout, scrollable=True, size=(800, 400))])

        #
        # TABS
        #

        self.cancel_layout = [sg.Button('Cancel', key="cancel")]

        self.layout = [
            self.cancel_layout,  # , sg.Button(' Next Card -> ', key="next")],
            [
                sg.TabGroup(
                    [
                        [
                            sg.Tab('Burn', burn_layout, key="panel-burn"),
                            sg.Tab('Log', log_scroll_layout, key="panel-lg", background_color='white'),
                            sg.Tab('Network', net_layout, key="panel-net", background_color='white'),
                            sg.Tab('Rack', rack_layout, key="panel-rack", background_color='white')
                        ]
                    ],
                    tooltip='Rack', key="mytabs")
            ]
        ]
        return self.layout

    def create_diag(self, name, _new=True):

        self.logger("Creating Diagrams")

        directory = path_expand("~/.cloudmesh/gui")
        Shell.mkdir(directory)
        location = f"{directory}/{name}"

        hostnames = Parameter.expand(self.hostnames)
        if _new:
            rack = Diagram(names=hostnames, name=hostnames[0])
            rack.save(location)

        diagram = Diagram()
        diagram.load(location)

        diagram.render_rack()
        diagram.saveas(f"{location}-rack", kind="rack", output="png")

        diagram.render_bridge_net()
        diagram.saveas(f"{location}-net", kind="net", output="png")

    def logger(self, msg, end="\n\n"):
        try:
            text = self.window['log']
            text.update(text.get() + msg + end)
            self.window.Refresh()
        except:  # noqa
            print(msg)

    def set_diagram_value(self, name, entry, attribute, value):
        directory = path_expand("~/.cloudmesh/gui")
        Shell.mkdir(directory)
        location = f"{directory}/{name}"
        diagram = Diagram()
        diagram.load(location)
        data = {
            attribute: value
        }
        diagram.set(entry, **data)
        diagram.save(location)

    def update_diagram_colors(self, cluster, host, color):

        self.set_diagram_value(cluster, host, "rack.color", color)
        self.set_diagram_value(cluster, host, "net.color", color)
        self.create_diag(cluster, _new=False)

        rack_file = f"~/.cloudmesh/gui/{cluster}-rack.png"
        net_file = f"~/.cloudmesh/gui/{cluster}-net.png"

        self.window['net-image'].update(data=image(net_file))
        self.window['rack-image'].update(data=image(rack_file))
        self.window.Refresh()

    def set_button_color(self, host, color):
        host = str(host)
        try:
            self.window.FindElement(f'button-{host}').Update(button_color=('white', color))
            self.window.Refresh()
        except:   # noqa
            pass

    def run(self):

        Sudo.password()

        host = None
        ips = None
        hostnames = None
        key = None
        event = None
        tags = None
        kind = None
        device = None

        while True:

            event, values = self.window.read()

            if event in ("Cancel", 'cancel', None):
                if not self.no_diagram:
                    rack_file = f"~/.cloudmesh/gui/{self.manager}-rack.png"
                    net_file = f"~/.cloudmesh/gui/{self.manager}-net.png"
                    os.remove(path_expand(rack_file))
                    os.remove(path_expand(net_file))
                break

            #
            # UPDATE OS SELECTION
            #
            if event.startswith("os"):
                for image in values:
                    if image.startswith("os") and values[image]:
                        break

                self.logger(f"Switch OS to: {image}")

                image_manager = image_tags[image]["manager"]
                image_worker = image_tags[image]["worker"]

                self.window[f'tags-{self.manager}'].update(image_manager)
                for worker in self.workers:
                    self.window[f'tags-{worker}'].update(image_worker)
                self.window.Refresh()

            if event.startswith("button"):

                #
                # set imaged string
                #
                imaged = values['imaged']
                if not imaged:
                    self.imaged_str = "--imaged"
                else:
                    self.imaged_str = ""
                #
                # handle device, hostnames and ips
                #
                ips = []
                hostnames = []
                tags = []
                for entry in values:
                    if str(entry).startswith("name"):
                        hostnames.append(values[entry])
                    if str(entry).startswith("ip"):
                        ips.append(values[entry])
                    if str(entry).startswith("device-") and values[entry]:
                        device = "/dev/" + entry.replace("device-", "")
                    if str(entry).startswith("tag") and values[entry]:
                        tags.append(values[entry])
                key = values['key']
                self.hostnames_str = ','.join(hostnames)
                self.ips_str = ','.join(ips)
                #
                # get the ssid
                #
                self.ssid = values['ssid']
                self.wifipassword = values['wifi']

                host = event.replace("button-", "")
                self.set_button_color(host, 'grey')
                if host == self.manager:
                    kind = "manager"
                else:
                    kind = "worker"

                print()
                print("Host:    ", host)
                print("IPs:     ", ips)
                print("Hostnames", hostnames)
                print("Ssid:    ", self.ssid)
                print("Key:     ", key)
                print("Event:   ", event)
                print("Tags:    ", tags)
                print("Kind:    ", kind)
                print("Device:  ", device)
                print("Format   ", imaged)
                print()

                # Call burn function for manager and workers
                self.logger(f"Burning {kind} {host}")
                # tags = values[f'tags-{host}']
                self.window[f'status-{host}'].update(' Burning ')

                if not self.no_diagram:
                    self.update_diagram_colors(self.manager, host, "blue")
                self.hostnames_str = ','.join(hostnames)
                self.ips_str = ','.join(ips)
                # command = f"cms burn cluster --device={device}" \
                #          f" --hostname={self.hostnames_str}" \
                #          f" --ssid={self.ssid}" \
                #          f" --wifipassword={self.wifipassword}" \
                #          f" --ip={self.ips_str}" \
                #          f" --burning={host}" \
                #          " -y" \
                #          f" {self.imaged_str}"

                manager, workers = Host.get_hostnames(hostnames)
                filename = path_expand(f"~/.cloudmesh/inventory-{manager}.yml")
                _build_default_inventory(filename=filename, manager=manager,
                                         workers=workers, ips=ips, images=tags)
                if host == manager:
                    command = f"cms burn raspberry {host}" \
                              f" --device={device}" \
                              f" --ssid={self.ssid}" \
                              f" --wifipassword={self.wifipassword}"
                else:
                    command = f"cms burn raspberry {host}" \
                              f" --device={device}"
                    print(command)

                try:
                    self.logger(f"Executing: {command}")
                    if self.dryrun:
                        time.sleep(0.5)
                    else:
                        os.system(command)
                    self.window[f'status-{host}'].update(' Completed ')
                    if not self.no_diagram:
                        self.update_diagram_colors(self.manager, host, "green")
                    self.set_button_color(host, 'green')

                except Exception as e:
                    print(e)
                    self.logger("Command failed")
                    if not self.no_diagram:
                        self.update_diagram_colors(self.manager, host, "orange")
                    self.set_button_color(host, 'red')

                self.window.FindElement(f'button-{host}').Update(button_color=('white', 'green'))

                self.window.Refresh()

        print('exit')
        self.window.close()



import textwrap


class Passwd:

    file = textwrap.dedent("""
        root:x:0:0:root:/root:/bin/bash
        daemon:x:1:1:daemon:/usr/sbin:/usr/sbin/nologin
        bin:x:2:2:bin:/bin:/usr/sbin/nologin
        sys:x:3:3:sys:/dev:/usr/sbin/nologin
        sync:x:4:65534:sync:/bin:/bin/sync
        games:x:5:60:games:/usr/games:/usr/sbin/nologin
        man:x:6:12:man:/var/cache/man:/usr/sbin/nologin
        lp:x:7:7:lp:/var/spool/lpd:/usr/sbin/nologin
        mail:x:8:8:mail:/var/mail:/usr/sbin/nologin
        news:x:9:9:news:/var/spool/news:/usr/sbin/nologin
        uucp:x:10:10:uucp:/var/spool/uucp:/usr/sbin/nologin
        proxy:x:13:13:proxy:/bin:/usr/sbin/nologin
        www-data:x:33:33:www-data:/var/www:/usr/sbin/nologin
        backup:x:34:34:backup:/var/backups:/usr/sbin/nologin
        list:x:38:38:Mailing List Manager:/var/list:/usr/sbin/nologin
        irc:x:39:39:ircd:/var/run/ircd:/usr/sbin/nologin
        gnats:x:41:41:Gnats Bug-Reporting System (admin):/var/lib/gnats:/usr/sbin/nologin
        nobody:x:65534:65534:nobody:/nonexistent:/usr/sbin/nologin
        systemd-timesync:x:100:102:systemd Time Synchronization,,,:/run/systemd:/usr/sbin/nologin
        systemd-network:x:101:103:systemd Network Management,,,:/run/systemd:/usr/sbin/nologin
        systemd-resolve:x:102:104:systemd Resolver,,,:/run/systemd:/usr/sbin/nologin
        _apt:x:103:65534::/nonexistent:/usr/sbin/nologin
        pi:x:1000:1000:,,,:/home/pi:/bin/bash
        messagebus:x:104:110::/nonexistent:/usr/sbin/nologin
        _rpc:x:105:65534::/run/rpcbind:/usr/sbin/nologin
        statd:x:106:65534::/var/lib/nfs:/usr/sbin/nologin
        sshd:x:107:65534::/run/sshd:/usr/sbin/nologin
        avahi:x:108:113:Avahi mDNS daemon,,,:/var/run/avahi-daemon:/usr/sbin/nologin
        lightdm:x:109:114:Light Display Manager:/var/lib/lightdm:/bin/false
        rtkit:x:110:116:RealtimeKit,,,:/proc:/usr/sbin/nologin
        pulse:x:111:119:PulseAudio daemon,,,:/var/run/pulse:/usr/sbin/nologin
        saned:x:112:122::/var/lib/saned:/usr/sbin/nologin
        hplip:x:113:7:HPLIP system user,,,:/var/run/hplip:/bin/false
        colord:x:114:123:colord colour management daemon,,,:/var/lib/colord:/usr/sbin/nologin
        """)

    shadow = textwrap.dedent("""
        root:*:18638:0:99999:7:::
        daemon:*:18638:0:99999:7:::
        bin:*:18638:0:99999:7:::
        sys:*:18638:0:99999:7:::
        sync:*:18638:0:99999:7:::
        games:*:18638:0:99999:7:::
        man:*:18638:0:99999:7:::
        lp:*:18638:0:99999:7:::
        mail:*:18638:0:99999:7:::
        news:*:18638:0:99999:7:::
        uucp:*:18638:0:99999:7:::
        proxy:*:18638:0:99999:7:::
        www-data:*:18638:0:99999:7:::
        backup:*:18638:0:99999:7:::
        list:*:18638:0:99999:7:::
        irc:*:18638:0:99999:7:::
        gnats:*:18638:0:99999:7:::
        nobody:*:18638:0:99999:7:::
        systemd-timesync:*:18638:0:99999:7:::
        systemd-network:*:18638:0:99999:7:::
        systemd-resolve:*:18638:0:99999:7:::
        _apt:*:18638:0:99999:7:::
        pi:$6$ZTD6hJH7f.ZyvE3M$Tgfv.ULQpXRtW7YRxImuFX1qrfVO5BGwTm17w/0.WYBiEck6mE5vRcFfJy.NgrYIJqU.aOGpu//hhYwL1uh8T.:18638:0:99999:7:::
        messagebus:*:18638:0:99999:7:::
        _rpc:*:18638:0:99999:7:::
        statd:*:18638:0:99999:7:::
        sshd:*:18638:0:99999:7:::
        avahi:*:18638:0:99999:7:::
        lightdm:*:18638:0:99999:7:::
        rtkit:*:18638:0:99999:7:::
        pulse:*:18638:0:99999:7:::
        saned:*:18638:0:99999:7:::
        hplip:*:18638:0:99999:7:::
        colord:*:18638:0:99999:7:::
        """)



import hashlib
import os
import platform
import sys

from cloudmesh.common.console import Console
from cloudmesh.common.util import readfile

# import mmap

BUF_SIZE = 65536


def sha1sum(filename=None):
    Console.info("Verifying sha1")
    h = hashlib.sha1()
    with open(filename, 'rb') as f:
        # with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as mm:
        #     h.update(mm)
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            h.update(data)
    return h.hexdigest()


def sha256sum(filename=None):
    Console.info("Verifying sha256")
    h = hashlib.sha256()
    with open(filename, 'rb') as f:
        # with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as mm:
        #     h.update(mm)
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            h.update(data)
    return h.hexdigest()


def os_is_windows():
    """
    Checks if the os is windows

    :return: True is windows
    :rtype: bool
    """
    return platform.system() == "Windows"


# noinspection PyBroadException
def os_is_linux():
    """
    Checks if the os is linux

    :return: True is linux
    :rtype: bool
    """
    try:
        content = readfile('/etc/os-release')
        return platform.system() == "Linux" and "raspbian" not in content
    except:  # noqa: E722
        return False


def os_is_mac():
    """
    Checks if the os is macOS

    :return: True is macOS
    :rtype: bool
    """
    return platform.system() == "Darwin"


# noinspection PyBroadException
def os_is_pi():
    """
    Checks if the os is Raspberry OS

    :return: True is Raspberry OS
    :rtype: bool
    """
    try:
        content = readfile('/etc/os-release')
        return platform.system() == "Linux" and "raspbian" in content
    except:  # noqa: E722
        return False


def check_root(dryrun=False, terminate=True):
    """
    check if I am the root user. If not simply exist the program.

    TODO: should probably move to cloudmesh.common

    :param dryrun: if set to true, does not terminate if not root user
    :type dryrun: bool
    :param terminate: terminates if not root user and dryrun is False
    :type terminate: bool
    """
    uid = os.getuid()
    if uid == 0:
        print("You are executing a a root user")
    else:
        print("You do not run as root")
        if terminate and not dryrun:
            sys.exit()



import crypt
import os
import random
import re
import string
import subprocess
import sys
import textwrap
import time
from getpass import getpass

from cloudmesh.bridge.Bridge import Bridge
from cloudmesh.burn.image import Image
from cloudmesh.burn.sdcard import SDCard
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_pi
from cloudmesh.burn.util import os_is_windows
from cloudmesh.burn.wifi.provider import Wifi
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.Host import Host
from cloudmesh.common.Shell import Shell
from cloudmesh.common.Shell import windows_not_supported
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.security import generate_strong_pass
from cloudmesh.common.sudo import Sudo
from cloudmesh.common.systeminfo import get_platform
from cloudmesh.common.util import banner
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.util import yn_choice
from cloudmesh.inventory.inventory import Inventory


# noinspection PyPep8
class Burner(object):

    def __init__(self):
        """
        Initializes the burner
        """
        self.hostname = None
        self.keypath = None

    # noinspection PyBroadException
    @windows_not_supported
    def check(self, device="/dev/sdX"):
        """
        This method checks what configurations are placed on the PI se card

        :param device:
        :type device:
        :return:
        :rtype:
        """

        data = {
            "ssh": None,
            "hostname": None,
            "ip": None,
            "wifi": None,
            "psk": None,
            "ssid": None,
            "auth_key": None
        }

        host = get_platform()

        if host == "windows":
            Console.error("Not yet implemented for this OS")
            return ""

        card = SDCard(host_os=host)
        # ssh

        try:
            data["ssh"] = \
                os.path.exists(f'{card.boot_volume}/ssh') or \
                os.path.exists(f'{card.root_volume}/etc/systemd/system/sshd.service')
        except Exception as e:
            data["ssh"] = str(e)

        # auth_key

        try:
            data["auth_key"] = \
                os.path.exists(f'{card.root_volume}/home/pi/.ssh/authorized_keys')
            if data["auth_key"]:
                content = readfile(f"{card.root_volume}/home/pi/.ssh/authorized_keys")
                data["auth_key"] = content.split()[-1]
        except Exception as e:
            data["auth_key"] = str(e)

        # hostname

        try:
            content = readfile(f"{card.root_volume}/etc/hostname").strip()
            data['hostname'] = content
        except Exception as e:
            data["hostname"] = str(e)

        # ip

        try:
            content = readfile(f"{card.root_volume}/etc/dhcpcd.conf")
            for line in content.splitlines():
                if line.startswith('static ip_address='):
                    data['ip'] = line[18:]
            if data['ip'] is None:
                data['ip'] = 'False'
        except Exception as e:
            data["ip"] = str(e)

        # wifi

        try:
            location = f"{card.boot_volume}/wpa_supplicant.conf"
            data["wifi"] = os.path.exists(location)

            if data["wifi"]:
                lines = readfile(location).splitlines()
                for line in lines:
                    for tag in ["ssid", "psk"]:
                        if f'{tag}=' in line:
                            data[tag] = line.split(f'{tag}=')[1].replace('"', "")

            else:
                data["wifi"] = False
                data["ssid"] = None
                data["psk"] = None

        except Exception as e:  # noqa: F841

            data["wifi"] = False
            data["ssid"] = None
            data["psk"] = None

        banner("Card Check")
        print(Printer.attribute(
            data,
            sort_keys=False,
            order=[
                "hostname",
                "ip",
                "ssh",
                "auth_key",
                "wifi",
                "psk",
                "ssid"
            ]
        ))

    @windows_not_supported
    def firmware(self, action="check"):
        """
        Checks or update the firmware
        :param action: the cations to be performed. It is "check" or "update"
        :type action: str
        """

        def _execute(command):
            print(command)
            print()
            os.system(command)
            print()

        if not os_is_pi():
            Console.error("This command can only be run on a PI")
        else:
            if action == "check":
                _execute('sudo apt update')
                _execute('sudo apt install rpi-eeprom')
                _execute("sudo rpi-eeprom-update")
                _execute("vcgencmd bootloader_version")
                print("* https://www.raspberrypi.org/documentation/hardware/raspberrypi/booteeprom.md")
                print()
            elif action == "update":
                _execute('sudo apt update')
                _execute('sudo apt install rpi-eeprom')
                _execute("sudo rpi-eeprom-update -a")
                os.system("sudo reboot")

    def mac(self, hostnames=None):
        """
        Sets the hostname for which we probe the MAC addresses

        :param hostnames: the hostnames
        :type hostnames: str
        """
        self.hostnames = hostnames

        print(hostnames)
        Console.error("Not yet implemented")
        return ""

    @windows_not_supported
    def set_locale(self, locale="en_US.UTF-8"):

        lang = textwrap.dedent(f'''
        LANG={locale}
        LC_ALL={locale}
        LANGUAGE={locale}
        #
        ''').strip()

        card = SDCard()

        # Write it 3 times as sometimes it does not work
        for i in range(0, 3):
            SDCard.writefile(f"{card.root_volume}/etc/default/locale", lang)

        locale_gen = SDCard.readfile(f"{card.root_volume}/etc/locale.gen",
                                     split=True, decode=True)
        for i in range(0, len(locale_gen)):
            if not locale_gen[i].startswith("#"):
                locale_gen[i] = "# " + locale_gen[i]
            if locale in locale_gen[i]:
                locale_gen[i] = locale_gen[i].replace("# ", "")
        locale_gen = "\n".join(locale_gen) + "\n"

        for i in range(0, 3):
            SDCard.writefile(f"{card.root_volume}/etc/locale.gen", locale_gen)

    def set_cmdline(self, cmdline):
        card = SDCard()
        filename = f'{card.boot_volume}/cmdline.txt'
        content = Sudo.readfile(filename=filename, split=False, decode=True)
        content = content.rstrip() + " " + cmdline
        SDCard.writefile(filename=f'{card.boot_volume}/cmdline.txt',
                         content=content)

    @windows_not_supported
    def set_hostname(self, hostname):
        """
        Sets the hostname on the sd card

        :param hostname: the hostname
        :type hostname: str
        """

        # On macOS we have
        # 644 -rw-r--r--  1 user  group  /Volumes/rootfs/etc/hostname

        # od -c  /Volumes/rootfs/etc/hostname
        # 0000000    r   a   s   p   b   e   r   r   y   p   i  \n
        # 0000014

        self.hostname = hostname
        card = SDCard()

        name = hostname.strip()

        # Write it 3 times as sometimes it does not work
        for i in range(0, 3):
            time.sleep(0.5)
            SDCard.writefile(f"{card.root_volume}/etc/hostname", name)

        # change last line of /etc/hosts to have the new hostname
        # 127.0.1.1 raspberrypi   # default
        # 127.0.1.1 red47         # new

        hosts = textwrap.dedent(f"""
        127.0.0.1  localhost
        ::1        localhost ip6-localhost ip6-loopback
        ff02::1    ip6-allnodes
        ff02::2    ip6-allrouters
        #
        127.0.1.1  {hostname}
        #
        """).strip()

        print(f'Writing: {card.root_volume}/etc/hosts')
        print(hosts)
        print()

        SDCard.writefile(f'{card.root_volume}/etc/hosts', hosts)

    def add_to_hosts(self, ip=None):
        hosts = SDCard.readfile('/etc/hosts', split=True, decode=True)

        replaced = False
        for i in range(len(hosts)):
            ip_host = hosts[i].split()

            if len(ip_host) > 1:
                if ip_host[0] == ip:
                    ip_host[1] = self.hostname
                    hosts[i] = f"{ip_host[0]}\t{ip_host[1]}"
                    replaced = True

                elif ip_host[1] == self.hostname:
                    ip_host[0] = ip
                    hosts[i] = f"{ip_host[0]}\t{ip_host[1]}"
                    replaced = True
        if not replaced:
            hosts.append(f"{ip}\t{self.hostname}")

        config = ""
        for line in hosts:
            config = config + line + '\n'

        SDCard.writefile('/etc/hosts', config + "\n")

    def write_cluster_hosts(self, cluster_hosts=None):
        card = SDCard()
        hosts = SDCard.readfile(f'{card.root_volume}/etc/hosts', split=False, decode=True)
        hosts = hosts + '\n'
        for ip, hostname in cluster_hosts:
            hosts = hosts + f"{ip}\t{hostname}\n"
        hosts = hosts + "#\n"
        SDCard.writefile(f'{card.root_volume}/etc/hosts', hosts)

    @windows_not_supported
    def set_static_ip(self,
                      ip=None,
                      iface="eth0",
                      mask="24",
                      router_ip="10.1.1.1",
                      write_local_hosts=True):
        """
        Sets the static ip on the sd card for the specified interface
        Also writes to manager hosts file for easy access

        :param ip: ips address
        :type ip: str
        :param iface: the network Interface
        :type iface: str
        :param mask: the subnet Mask
        :type mask: str
        :param router_ip:
        :type router_ip: str
        :param write_local_hosts:
        :type write_local_hosts: bool
        :return:
        :rtype:
        """
        # TODO:
        # router_ip statically set to default ip configured with cms bridge
        # create. Rewrite to consider the IP of the manager on iface

        card = SDCard()

        mountpoint = card.root_volume
        if self.hostname is not None and write_local_hosts:
            self.add_to_hosts(ip)

        iface = f'interface {iface}'
        static_ip = f'static ip_address={ip}/{mask}'
        static_routers = f'static routers={router_ip}'

        curr_config = SDCard.readfile(f'{mountpoint}/etc/dhcpcd.conf', decode=True, split=True)
        if iface in curr_config:
            Console.warning("Found previous settings. Overwriting")
            # If setting already present, replace it and the static ip line
            index = curr_config.index(iface)
            try:
                if 'static ip_address' not in curr_config[index + 1]:
                    Console.warning(
                        "Missing static ip_address assignment. Overwriting line")
                curr_config[index + 1] = static_ip

            except IndexError:
                Console.error(f'{mountpoint}/etc/dhcpcd.conf ends abruptly. Aborting')
                sys.exit(1)

        else:
            curr_config.append(iface)
            curr_config.append(static_ip)
            if router_ip:
                curr_config.append(static_routers)
            curr_config.append('\n')
            # curr_config.append('nolink\n')

        SDCard.writefile(f'{mountpoint}/etc/dhcpcd.conf', '\n'.join(curr_config))

    @windows_not_supported
    def keyboard(self, country="US"):
        """
        Sets the country on the SDCard

        :param country: two letter country code
        :type country: str
        :return: true if set
        :rtype: bool
        """

        # cat /etc/default/keyboard
        # # KEYBOARD CONFIGURATION FILE
        #
        # # Consult the keyboard(5) manual page.
        #
        # XKBMODEL="pc105"
        # XKBLAYOUT=us
        # XKBVARIANT=""
        # XKBOPTIONS=""
        #
        # BACKSPACE="guess"

        card = SDCard()

        layout = f"{card.root_volume}/etc/default/keyboard"

        content = SDCard.readfile(layout, decode=True, split=True)

        country = country.lower()
        found = False
        for i in range(0, len(content)):
            if content[i].strip().startswith("XKBLAYOUT"):
                content[i] = f"XKBLAYOUT={country}"
                found = True
                break

        content = "\n".join(content)
        print(content)
        SDCard.writefile(layout, content)
        Sudo.execute("sync")
        return found

    @windows_not_supported
    def set_key(self, key_file=None):
        """
        Copies the public key into the .ssh/authorized_keys file on the sd card

        :param key_file: key_file of public key, e.g. 'id_rsa' for ~/.ssh/id_rsa.pub
        :type key_file: str
        """
        # copy file on burner computer ~/.ssh/id_rsa.pub into
        #   mountpoint/home/pi/.ssh/authorized_keys

        card = SDCard()

        location = path_expand(key_file)

        SDCard.execute(f'mkdir -p {card.root_volume}/home/pi/.ssh/')
        SDCard.execute(f'cp {location} {card.root_volume}/home/pi/.ssh/authorized_keys')
        # cleanup
        if os.path.exists(f"{card.root_volume}/home/pi/.ssh/._authorized_keys"):
            SDCard.execute(f"rm -f {card.root_volume}/home/pi/.ssh/._authorized_keys")

    def write_fix(self, locale="en_US.UTF-8"):
        """
        Not yet integrated:
            /etc/systemd/system/sshd.service
            /etc/passwd
            /etc/ssh/sshd_config
        """
        timezone = Shell.timezone()
        script = textwrap.dedent(f"""
            #! /usr/bin/env python
            # file, owner, group, permissions
            import os
            if not os.path.exists("/boot/fixed"):
                files = [
                    ["/etc/default/keyboard", 0, 0, 0o644],
                    ["/boot/fix_permissions.py", 0, 0, 0o777],
                    ["/boot/ssh", 0, 0, 0o777],
                    ["/boot/wpa_supplicant.conf", 0, 0, 0o600],
                    ["/etc/hosts", 0, 0, 0o644],
                    ["/etc/default/locale", 0, 0, 0o644],
                    ["/etc/environment", 0, 0, 0o644],
                    ["/etc/locale.conf", 0, 0, 0o644],
                    ["/etc/passwd", 0, 0, 0o644],
                    ["/etc/dhcpcd.conf", 0, 109, 0o664],
                    ["/etc/hostname", 0, 0, 0o644],
                    ["/etc/ssh/sshd_config", 0, 0, 0o644],
                    ["/etc/shadow", 0, 42, 0o640],
                    ["/etc/rc.local", 0, 0, 0o751],
                    ["/home/pi/.ssh", 1000, 1000, 0o700],
                    ["/home/pi/.ssh/authorized_keys", 1000, 1000, 0o644],
                    ["/home/pi/.ssh/id_rsa", 1000, 1000, 0o600],
                    ["/home/pi/.ssh/id_rsa.pub", 1000, 1000, 0o644]
                ]
                for name, uid, guid, permission in files:
                    if os.path.exists(name):
                        os.chown(name, uid, guid)
                        os.chmod(name, permission)
                os.system("locale-gen {locale}")
                name = "/boot/fixed"
                os.system("touch /boot/fixed")
                os.chown("/boot/fixed", 0, 0)
                os.chmod("/boot/fixed", 0o644)
                os.system("sudo timedatectl set-timezone {timezone}")
                os.system("sudo sync")
                # os.system("sleep 90")
                # os.system("sudo reboot")
                #
        """)

        card = SDCard()
        fix = "/boot/fix_permissions.py"
        fix_on_sdcard = f"{card.boot_volume}/fix_permissions.py"
        SDCard.writefile(fix_on_sdcard, script)

        rc_local = f"{card.root_volume}/etc/rc.local"
        content = readfile(rc_local)
        if fix in content:
            return
        else:
            content = content.replace("exit 0", f"sudo python {fix}")
            content = content + "\n" + "exit 0\n"
            SDCard.writefile(rc_local, content)

    @windows_not_supported
    def enable_ssh(self):
        """
        Enables ssh on next boot of sd card
        """
        host = get_platform()

        if host in ['raspberry', "linux"]:
            sudo = True
        else:
            sudo = False

        card = SDCard(card_os="raspberry", host_os=host)
        if sudo:
            Sudo.password()
            command = f'sudo touch {card.boot_volume}/ssh'
        else:
            command = f'touch {card.boot_volume}/ssh'

        os.system(command)

        return ""

    # IMPROVE

    # TODO: docstring
    @windows_not_supported
    def disable_password_ssh(self):

        # sshd_config = self.filename("/etc/ssh/sshd_config")
        card = SDCard()
        sshd_config = f'{card.root_volume}/etc/ssh/sshd_config'
        new_sshd_config = ""
        updated_params = False

        def sets_param(param, line):
            """See if a config line sets this parameter to something."""
            # front is only whitespace maybe a comment
            front = r'^\s*#?\s*'
            # only whitespace between param and value
            middle = r'\s+'
            # end can include a comment
            end = r'\s*(?:#.*)?$'
            re_sets_param = front + param + middle + r'.*' + end
            return re.search(re_sets_param, line) is not None

        force_params = [
            ("ChallengeResponseAuthentication", "no"),
            ("PasswordAuthentication", "no"),
            ("UsePAM", "no"),
            ("PermitRootLogin", "no"),
        ]

        found_params = set()
        # with open(sshd_config, 'r') as f:
        f = SDCard.readfile(sshd_config, decode=True, split=True)

        for line in f:
            found_a_param = False
            for param, value in force_params:
                if sets_param(param, line):
                    # Only set the parameter once
                    if param not in found_params:
                        new_sshd_config += param + " " + value + "\n"
                        updated_params = True
                    found_a_param = True
                    found_params.add(param)
            if not found_a_param:
                new_sshd_config += line
        # Check if any params not found
        for param, value in force_params:
            if param not in found_params:
                new_sshd_config += param + " " + value + "\n"
                updated_params = True
        if updated_params:
            # NOTE: This is actually necessary, see comment in method
            #
            # as we do it on osx, we need to identify if this is still needed
            #
            # self.truncate_file(sshd_config)
            # with open(sshd_config, "w") as f:
            #     f.write(new_sshd_config)

            SDCard.writefile(sshd_config, new_sshd_config)

    @windows_not_supported
    def configure_wifi(self,
                       ssid=None,
                       psk=None,
                       card_os='raspberry',
                       host=None,
                       country=None):
        """
        Sets the wifi. Only works for psk based wifi

        :param ssid: the ssid
        :type ssid: str
        :param psk: the psk
        :type psk: str
        :param card_os: the os on the sdcard
        :type card_os: str
        :param host: the machine os running the command
        :type host: str
        :param country: two digit country code for rf settings
        :type country: str
        """

        country = country or 'US'

        card = SDCard(card_os=card_os, host_os=host)
        path = f"{card.boot_volume}/wpa_supplicant.conf"

        card_os = "raspberry"  # needs to become a parameter based on tag

        # noinspection PyPep8Naming
        WifiClass = Wifi(card_os=card_os)

        if card_os == "raspberry":
            if psk:
                if os_is_mac():
                    WifiClass.set(ssid=ssid, password=psk, country=country, location=path)
                else:
                    WifiClass.set(ssid=ssid, password=psk, country=country, location=path, sudo=True)
            else:
                if os_is_mac():
                    WifiClass.set(ssid=ssid, psk=False, country=country, location=path)
                else:
                    WifiClass.set(ssid=ssid, psk=False, country=country, location=path, sudo=True)
        else:
            if os_is_mac():
                WifiClass.set(ssid=ssid, password=psk, country=country, location=path)
            else:
                WifiClass.set(ssid=ssid, password=psk, country=country, location=path, sudo=True)

        return ""

    # This is to prevent desktop access of th pie (directly plugging monitor, keyboard, mouse into pi, etc.)
    #
    # Currently, ssh login is only possible with an authorized key. (No passwords)
    # Plugging pi directly into desktop, however, will still prompt for a user and password.
    # I can't figure out how to disable it

    @windows_not_supported
    def disable_terminal_login(self, mountpoint=None, password=None):
        """
        disables and replaces the password with a random string so that by
        accident the pi can not be logged into. The only way to login is via the
        ssh key

        :param mountpoint: the mountpount for the system
        :type mountpoint: str
        :param password: the password for login
        :type password: str
        :return: file in /etc/shadow
        :rtype: a written file
        """

        # Generates random salt for password generation
        def random_salt(length=10):
            letters_and_digits = string.ascii_letters + string.digits
            return ''.join(
                random.choice(letters_and_digits) for i in range(length))

        salt = random_salt()
        if password is not None:
            pswd = crypt.crypt(password, f'$6${salt}')
        else:
            raise NotImplementedError()

        # Make sure there's an 'x' in /etc/passwd
        #        with open(f'{mountpoint}/etc/passwd', 'r') as f:
        #            info = [l for l in f.readlines()]

        info = SDCard.readfile(f'{mountpoint}/etc/passwd', split=True, decode=True)

        for i in range(len(info)):
            inf = info[i].split(":")
            if inf[0] == 'pi':
                inf[1] = 'x'
                info[i] = ':'.join(inf)

        content = '\n'.join(info)

        # with open(f'{mountpoint}/etc/passwd', 'w') as f:
        #     f.writelines(info)

        SDCard.writefile(f'{mountpoint}/etc/passwd', content)

        # Add it to shadow file
        # with open(f'{mountpoint}/etc/shadow', 'r') as f:
        #     data = [l for l in f.readlines()]

        data = SDCard.readfile(f'{mountpoint}/etc/shadow', decode=True, split=True)

        content = ""
        for i in range(len(data)):
            dat = data[i].split(":")
            if dat[0] == 'pi':
                dat[1] = pswd
                data[i] = ':'.join(dat)

        content = '\n'.join(data)

        SDCard.writefile(f'{mountpoint}/etc/shadow', content)

    def generate_key(self, hostname=None):
        card = SDCard()
        # TODO investigate what happens if not run as UID 1000 (e.g. first user)

        cmd = f'mkdir -p {card.root_volume}/home/pi/.ssh/'
        os.system(cmd)

        cmd = f'ssh-keygen -q -N "" -C "pi@{hostname}" -f {card.root_volume}/home/pi/.ssh/id_rsa'
        os.system(cmd)

    @staticmethod
    def store_public_key():
        card = SDCard()
        cmd = f'cp {card.root_volume}/home/pi/.ssh/id_rsa.pub ~/.cloudmesh/cmburn/'
        os.system(cmd)

    @staticmethod
    def remove_public_key():
        cmd = 'rm -f ~/.cloudmesh/cmburn/id_rsa.pub'
        os.system(cmd)

    @staticmethod
    def get_tag(worker=True, card_os="raspberry"):
        tag = None
        if "raspberry" in card_os and worker:
            tag = "latest-lite"
        elif "raspberry" in card_os and not worker:
            tag = "latest-full"
        elif "ubuntu" in card_os and worker:
            tag = "ubuntu-20.04.2-64-bit"
        elif "ubuntu" in card_os and not worker:
            tag = "ubuntu-desktop"
        else:
            Console.error("For now only raspberry and ubuntu are supported")
        return tag

    def cluster(self, arguments=None):

        # is true when
        #
        # cms burn cluster --hostname=red,red00[1-2]
        #                  --device=/dev/sdb
        #                  --ip=10.1.1.[1-3]
        #                  --ssid=myssid
        #                  --wifipassword=mypass
        #

        Sudo.password()
        yes = arguments.yes
        arguments.os = arguments.os or "raspberry"

        if os_is_windows():
            Console.error("Only supported on Pi and Linux. On Mac you will "
                          "need to have ext4 write access.")
            return ""

        hostnames = Parameter.expand(arguments.hostname)
        manager, workers = Host.get_hostnames(hostnames)
        if arguments.burning is None:
            burning = hostnames
        else:
            burning = Parameter.expand(arguments.burning)

        if not (arguments.cluster and  # noqa: W504
                arguments.device and  # noqa: W504
                arguments.hostname):  # noqa: W504
            Console.error("Parameters not complete")
            return ""

        passwd = None
        if arguments.set_passwd:
            passwd = getpass("Host Password: ")
        elif "PASSWD" in os.environ:
            passwd = os.environ["PASSWD"]
        else:
            passwd = generate_strong_pass()

        if manager is not None and arguments.ssid is None:
            Console.error("Please set ssid")
            return ""

        if manager is not None and \
            (arguments.wifipassword is None
             or arguments.wifipassword.lower() in ['input', "none", ""]):  # noqa: W503
            print()
            arguments.wifipassword = getpass("Wifi Password: ")

        if workers is None:
            n = 1
        else:
            n = len(workers) + 1
        if arguments.ip is None:
            ips = Parameter.expand(f"10.1.1.[1-{n}]")
        else:
            ips = Parameter.expand(arguments.ip)

        cluster_hosts = tuple(zip(ips, hostnames))

        key = path_expand("~/.ssh/id_rsa.pub")

        banner("Parameters", figlet=True)

        print("Burning:      ", burning)
        print("Manager:      ", manager)
        print("Workers:      ", workers)
        print("IPS:          ", ips)
        print("Device:       ", arguments.device)
        print("SSID:         ", arguments.ssid)
        print("Wifi Password:", "********")
        print("Key:          ", key)
        print("Blocksize:    ", arguments.bs)
        print("OS:           ", arguments.os)
        print("Imaged:       ", arguments.imaged)
        print("Host Password:", "********")

        if not (yes or yn_choice('\nWould you like to continue?')):
            Console.error("Aborting ...")
            return ""

        banner("Download Images", figlet=True)

        StopWatch.start("download image")

        result = Image.create_version_cache()
        if result is None:
            result = Image.create_version_cache(refresh=True)

        image = Image()
        image.read_version_cache()

        if "raspberry" in arguments.os:
            if workers is not None:
                image.fetch(tag=["latest-lite"])
            if manager is not None:
                image.fetch(tag=["latest-full"])
        elif "ubuntu" in arguments.os:
            if workers is not None:
                image.fetch(tag=["ubuntu-20.04.2-64-bit"])
            if manager is not None:
                image.fetch(tag=["ubuntu-desktop"])
        else:
            Console.error("For now only raspberry and ubuntu are supported")
        StopWatch.stop("download image")
        StopWatch.status("download image", True)

        multi = MultiBurner()

        if manager is not None and manager in burning:
            banner("Burn the Manager", figlet=True)

            Console.info(f"Please insert the SD Card: {manager}")

            Console.info(f"Preparing to burn the manager: {manager}")
            if not yn_choice('\nWould you like to continue?'):
                Console.error("Aborting ...")
                return ""

            multi.burn(device=arguments.device,
                       blocksize=arguments.bs,
                       progress=True,
                       hostname=manager,
                       ip=ips[0],
                       key=key,
                       password=passwd,
                       ssid=arguments.ssid,
                       psk=arguments.wifipassword,
                       formatting=not arguments.imaged,
                       imaging=not arguments.imaged,
                       tag=Burner.get_tag(card_os=arguments.os, worker=False),
                       router=None,
                       generate_key=True,
                       store_key=True,
                       write_local_hosts=False,
                       cluster_hosts=cluster_hosts,
                       yes=yes)

            Console.info(f"Completed manager: {manager}")

        if workers is not None:
            banner("Burn the Workers", figlet=True)

            Console.info(f"Preparing to burn the workers: {workers}")
            for worker, ip in tuple(zip(workers, ips[1:])):
                if worker not in burning:
                    continue

                banner(f"Worker {worker}", figlet=True)

                print()
                Console.info(f"Please insert SD Card for worker {worker}")
                print()

                if not yn_choice("Say Y once you inserted it. Press no to terminate ..."):
                    return ""

                multi.burn(device=arguments.device,
                           blocksize=arguments.bs,
                           progress=True,
                           hostname=worker,
                           ip=ip,
                           key='~/.cloudmesh/cmburn/id_rsa.pub',
                           password=passwd,
                           ssid=None,
                           psk=None,
                           formatting=not arguments.imaged,
                           imaging=not arguments.imaged,
                           tag=Burner.get_tag(card_os=arguments.os, worker=True),
                           router=ips[0],
                           generate_key=False,
                           store_key=False,
                           write_local_hosts=False,
                           cluster_hosts=cluster_hosts,
                           yes=yes)

            Console.info(f"Completed workers: {workers}")
            Console.info("Cluster burn is complete.")
            # Burner.remove_public_key()
            # TODO remove public key on GUI close or single cluster cmd complete

        banner("Benchmark", figlet=True)

        StopWatch.status("load", True)
        StopWatch.status("command", True)

        Benchmark.print(sysinfo=True, csv=True, tag="local")
        banner("Done", figlet=True)
        return ""


class MultiBurner(object):
    """
    This class uses a single or multicard burner to burn SD Cards. It detects
    how many SD Cards are there and uses them. We assume no other USB devices
    are plugged in other than a keyboard or a mouse.

    """

    # System command that uses subprocess to execute terminal commands
    # Returns the stdout of the command
    def system_exec(self, command=None):
        """

        :param command:
        :return:
        """
        res = subprocess.getstatusoutput(command)
        # If exit code is not 0, warn user
        if res[0] != 0:
            Console.warning(
                f'Warning: "{command}" did not execute properly -> {res[1]} :: exit code {res[0]}')

        return res[1]

    # noinspection PyUnboundLocalVariable
    def burn_all(self,
                 burning=None,
                 image="latest",
                 device=None,
                 blocksize="4M",
                 progress=True,
                 hostnames=None,
                 ips=None,
                 key=None,
                 password=None,
                 ssid=None,
                 psk=None,
                 formatting=True,
                 imaging=True,
                 tag='latest-lite',
                 locale="en_US.UTF-8",
                 yes=False):
        """
        TODO: provide documentation

        :param image:
        :type image:
        :param device:
        :type device:
        :param blocksize:
        :type blocksize:
        :param progress:
        :type progress:
        :param hostnames:
        :type hostnames:
        :param ips:
        :type ips:
        :param key:
        :type key:
        :param password:
        :type password:
        :param ssid:
        :type ssid:
        :param psk:
        :type psk:
        :param tag:
        :type tag:
        :param locale:
        :type locale:
        :param formatting:
        :type formatting:
        :param yes:
        :type yes:
        :return:
        :rtype:
        """

        # :param devices: string with device letters

        # print (device)

        #
        # define the dev
        #
        devices = {}  # dict of {device_name: empty_status}
        #
        # probe the dev
        #
        card = SDCard()
        info_statuses = card.info()

        # If the user specifies a particular device, we only care about that
        # device

        if device is not None:
            for dev in device:
                devices[dev] = info_statuses[dev]['empty']
            # Change to empty to skip next loop
            info_statuses = {}

        for device in info_statuses.keys():
            # print("call the info command on the device and "
            #      "figure out if an empty card is in it")
            # change the status based on what you found
            devices[device] = info_statuses[device]['empty']

        # if we detect a non empty card we interrupt and tell
        # which is not empty.
        # (print out status of the devices in a table)
        device_statuses = devices.values()
        if False in device_statuses:
            print("\nEmpty status of devices:")
            for dev, empty_status in devices.items():
                x = "" if empty_status else "not "
                Console.info(f"Device {dev} is {x}empty")
        print()

        # detect if there is an issue with the cards, readers
        # TODO what exactly should be done here?

        # TODO This does nothing relevant
        # ask if this is ok to burn otherwise
        # burn_all = yn_choice("Format the card before burning?")

        # # if no burn all of them for which we have status "empty card"
        # if not burn_all:
        #     # delete from devices dict any non-empty devices
        #     devices_to_delete = []
        #     for device in devices.keys():
        #         if devices[device]:
        #             # can't delete while iterating
        #             devices_to_delete.append(device)
        #     for device in devices_to_delete:
        #         del devices[device]

        print("Burning these devices:")
        print(' '.join(devices.keys()))

        if burning is None:
            burning = hostnames

        keys = list(devices.keys())
        count = 0
        for i in range(len(hostnames)):
            # for device, status in devices.items():
            # We might be using one device slot to burn multiple cards
            device = keys[i % len(keys)]
            # status = devices[device]
            hostname = hostnames[i]
            ip = None if not ips else ips[i]

            self.burn(image=image,
                      device=device,
                      blocksize=blocksize,
                      progress=progress,
                      hostname=hostname,
                      ip=ip,
                      key=key,
                      password=password,
                      ssid=ssid,
                      psk=psk,
                      formatting=formatting,
                      imaging=imaging,
                      tag=tag,
                      locale=locale,
                      yes=yes)

            count += 1
            Console.info(f'Burned card {count}')
            print()
            Console.info('Please remove the card')
            print()
            os.system('tput bel')  # ring the terminal bell to notify user
            if i < len(hostnames) - 1:
                if (i + 1) != ((i + 1) % len(keys)):
                    slot = keys[(i + 1) % len(keys)]
                    print()
                    print(f"Please remove any card from slot {slot} and insert a new one.")
                    if yn_choice("Is the card inserted and do you wish to continue?"):
                        pass
                    else:
                        return ""

                print('Burning next card...')
                print()

        Console.info(f"You burned {count} SD Cards")
        Console.ok("Done :)")

    # noinspection PyBroadException
    def burn(self,
             image="latest",
             device=None,
             blocksize="4M",
             progress=True,
             hostname=None,
             ip=None,
             key=None,
             password=None,
             ssid=None,
             psk=None,
             formatting=True,
             imaging=True,
             tag='latest-lite',
             router="10.1.1.1",
             generate_key=False,
             store_key=False,
             write_local_hosts=True,
             cluster_hosts=None,
             keyboard="us",
             locale="en_US.UTF-8",
             yes=False):
        """
        Burns the image on the specific device

        TODO: provide documentation

        :param image:
        :type image:
        :param device:
        :type device:
        :param blocksize:
        :type blocksize:
        :type blocksize:
        :param progress:
        :type progress:
        :param hostname: The hostnames to burn
        :type hostname: str
        :param ip:
        :type ip:
        :param key:
        :type key:
        :param password:
        :type password:
        :param ssid:
        :type ssid:
        :param psk:
        :type psk:
        :param formatting:
        :type formatting:
        :param yes:
        :type yes:
        :param tag:
        :type tag:
        :param locale:
        :type locale:
        :param keyboard:
        :type keyboard:
        :param router:
        :type router:
        :param cluster_hosts:
        :type cluster_hosts:
        :param generate_key:
        :type generate_key:
        :param store_key:
        :type store_key:
        :param write_local_hosts:
        :type write_local_hosts:
        :return:
        :rtype:
        """
        # TODO
        card = SDCard()
        # boot_volume = card.boot_volume
        root_volume = card.root_volume
        if key is None:
            key = '~/.ssh/id_rsa.pub'

        elif key == 'root':
            key = f'/{key}/.ssh/id_rsa.pub'

        # don't do the input() after burning the last card
        # use a counter to check this

        counter = 0
        burner = Burner()

        print("counter", counter)
        StopWatch.start(f"create {device} {hostname}")

        card = SDCard()
        if formatting:
            StopWatch.start(f"format {device} {hostname}")
            success = card.format_device(device=device,
                                         unmount=True,
                                         yes=yes)
            StopWatch.stop(f"format {device} {hostname}")

            if not success:
                Console.warning("Skipping card due to failed format. "
                                "Continuing with next hostname.")
                return
            StopWatch.status(f"format {device} {hostname}", True)

        if os_is_linux() or os_is_pi():
            card.unmount(device=device)  # can not fully eject before burn on pi or linux
        elif os_is_mac():
            card.unmount(device=device)

        if imaging:
            StopWatch.start(f"write image {device} {hostname}")
            card.burn_sdcard(tag=tag,
                             device=device,
                             blocksize=blocksize,
                             name=hostname,
                             yes=yes)
            StopWatch.stop(f"write image {device} {hostname}")
            StopWatch.status(f"write image {device} {hostname}", True)

        if os_is_linux():
            # full unmount is needed before mount on linux
            # added here for --imaged option to work
            card.unmount(device=device, full=True)

        Sudo.password()
        StopWatch.start(f"write host data {device} {hostname}")
        card.mount(device=device)
        Sudo.execute("sync")

        burner.keyboard(country=keyboard)
        burner.set_hostname(hostname)
        burner.set_locale(locale=locale)
        if generate_key:
            burner.generate_key(hostname)
        if store_key:
            Burner.store_public_key()
        burner.disable_terminal_login(root_volume, password)
        if ssid:
            Console.warning("In the future, try to interface with the workers via "
                            "ethernet/switch rather than WiFi")
            burner.configure_wifi(ssid, psk)
        burner.enable_ssh()
        burner.disable_password_ssh()
        burner.set_key(key)
        if ip:
            # TODO
            # interface = "wlan0" if ssid is not None else "eth0"
            #
            # we can't rely on ssid to determine which interface to static set.
            # for burning a cluster we want to static set the eth on manager
            # and enable wifi. We will need to add options --wifi_ip
            # --wifi_router
            #
            interface = 'eth0'
            burner.set_static_ip(ip, iface=interface, router_ip=router,
                                 write_local_hosts=write_local_hosts)
            if not write_local_hosts:
                burner.write_cluster_hosts(cluster_hosts)
        burner.write_fix()
        try:
            os.system(f"sudo rm -f {card.root_volume}/etc/xdg/autostart/piwiz.desktop")
        except:  # noqa: E722
            Console.error("Gui wizzard not found at "
                          f"{card.root_volume}/etc/xdg/autostart/piwiz.desktop")

        card.unmount(device=device, full=True)
        StopWatch.stop(f"write host data {device} {hostname}")
        StopWatch.status(f"write host data {device} {hostname}", True)

        time.sleep(2)
        StopWatch.stop(f"create {device} {hostname}")
        StopWatch.status(f"create {device} {hostname}", True)

    def burn_inventory(self,
                       inventory=None,
                       name=None,
                       device=None,
                       locale='en_US.UTF-8',
                       yes=False,
                       passwd=None):
        banner("Burning Inventory", figlet=True)
        i = Inventory(inventory)
        i.print()

        name = Parameter.expand(name)
        manager, workers = Host.get_hostnames(name)

        # name formatted as manager,worker
        # if ',' in name:
        #    manager, worker = name.split(',')
        #    workers = Parameter.expand(worker)
        # else:

        devices = Parameter.expand(device)

        # Warn user if they are burning non-empty devices
        card = SDCard()
        info_statuses = card.info(print_stdout=False)

        if os_is_mac():
            Console.warning(" ignoring device check")
            pass
        else:
            try:
                empty_statuses = {}
                for device in devices:
                    empty_statuses[device] = info_statuses[device]['empty']
            except KeyError:
                Console.error("Device specified not found by cms burn info. "
                              "Did you specify the correct path? Is the card properly inserted?")
                return

            for dev, empty_status in empty_statuses.items():
                if not empty_status:
                    if not (yes or yn_choice(f"Device {dev} is not empty. Do you wish to continue?")):
                        Console.error("Terminating")
                        return

        if manager is not None:
            manager_search_results = i.find(host=manager)
            if len(manager_search_results) == 0:
                Console.error(f"Could not find {manager} in inventory {inventory}. "
                              "Please correct before continuing.")
                return
            elif len(manager_search_results) > 1:
                Console.error(
                    f"Found duplicate {manager} configurations in inventory {inventory}. "
                    "Please correct before contuing")
                return

        if manager is not None:
            manager_config = {
                "hostname": manager,
                "tag": i.get(manager, "tag"),
                "ip": i.get(manager, "ip"),
                "services": i.get(manager, "services"),
                "keyfile": i.get(manager, "keyfile"),
                "dns": ','.join(i.get(manager, "dns"))
            }
        else:
            manager_config = None

        worker_configs = []
        if workers:
            for worker in workers:
                worker_config = {
                    "hostname": worker,
                    "tag": i.get(worker, "tag"),
                    "ip": i.get(worker, "ip"),
                    "services": i.get(worker, "services"),
                    "keyfile": i.get(worker, "keyfile"),
                    "router": i.get(worker, "router")
                }
                worker_configs.append(worker_config)

        _, system_hostname = subprocess.getstatusoutput("cat /etc/hostname")

        banner("Retrieving Images", figlet=True)
        # Get Images if not already downloaded
        result = Image.create_version_cache()
        if result is None:
            result = Image.create_version_cache(refresh=True)

        image = Image()
        image.read_version_cache()

        if manager is not None and system_hostname != manager_config["hostname"]:
            image.fetch(tag=manager_config["tag"])
        if workers is not None:
            image.fetch(tag=worker_configs[0]["tag"])

        # Set up this pi as a bridge if the hostname is the same
        # as the manager and if the user wishes
        if manager is not None:
            if system_hostname == manager_config["hostname"]:
                if yes or yn_choice("Manager hostname is the same as this system's "
                                    "hostname. Is this intended?"):
                    if yes or yn_choice("Do you wish to configure this system as a WiFi "
                                        "bridge? A restart is required after this "
                                        "command terminates"):
                        Bridge.create(managerIP=manager_config['ip'],
                                      priv_interface='eth0',
                                      ext_interface='wlan0',
                                      dns=manager_config["dns"])
                else:
                    Console.error("Terminating")
                    return
            else:
                self.burn(
                    device=device,
                    hostname=manager_config["hostname"],
                    ip=manager_config["ip"],
                    key=manager_config["keyfile"],
                    tag=manager_config["tag"],
                    password=passwd or generate_strong_pass(),
                    locale=locale)
                print("Safe to remove manager card.")
                if not workers or not yn_choice(f"Insert new card into slot {device}."
                                                " Is it inserted and do you wish to continue?"):
                    Console.ok("Done.")

        # The code below was taken from self.multi_burn
        # It would be nice to move this functionality of cycling over
        # sd card slots into a nice one liner function
        count = 0
        for i in range(len(worker_configs)):

            device = devices[i % len(devices)]
            worker_config = worker_configs[i]
            self.burn(
                device=device,
                hostname=worker_config["hostname"],
                ip=worker_config["ip"],
                key=worker_config["keyfile"],
                tag=worker_config["tag"],
                password=passwd or generate_strong_pass(),
                router=worker_config["router"] or (manager_config["ip"] if manager_config else None),
                locale=locale)

            count += 1
            Console.info(f'Burned card {count}')
            print()
            Console.info('Card is safe to remove')
            print()

            if i < len(worker_configs) - 1:
                if (i + 1) != ((i + 1) % len(devices)):
                    slot = devices[(i + 1) % len(devices)]
                    print()
                    print(f"Please remove any card from slot {slot} and insert a new one.")
                    if yn_choice("Is the card inserted and do you wish to continue?"):
                        pass
                    else:
                        return ""

        Console.info(f"You burned {count} SD Cards")
        Console.ok("Done :)")



import oyaml as yaml

from cloudmesh.common.console import Console
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand, writefile


class Userdata:
    """
    A Builder class for cloud-config Userdata

    A Builder class follows the Builder design pattern. This design pattern allows
    users of the class to construct a complex data structure using purely instance
    methods. This is very useful in that the user of the class need not concern
    themselves with the underlying implementation of the class. Consequentially,
    changes to this class can be made without needing a refactor of all existing calls.

    Examples:

    d = Userdata()\
    .with_ssh_password_login()\
    .with_locale()\
    .with_hostname(hostname='testserver')\
    .with_default_user()\
    .with_authorized_keys(key=['ssh-rsa AAAA.....user@laptop'])\
    .with_set_wifi_country(country='US')\
    .with_write_files(path='/home/ubuntu/.ssh/id_rsa', content='this is a ssh key', permissions='0600')\
    .with_write_files(path='/home/ubuntu/.ssh/id_rsa.pub', content='this is a pub key', permissions='0644')\
    .with_fix_user_dir_owner(user='ubuntu')

    print(d)

    #cloud-config
    ssh_pwauth: yes
    locale: en_US
    preserve_hostname: false
    hostname: testserver
    chpasswd:
    expire: true
    list:
    - ubuntu: ubuntu
    ssh_authorized_keys:
    - ssh-rsa AAAA.....user@laptop
    bootcmd:
    - - sudo
      - iw
      - reg
      - set
      - US
    runcmd:
    - - sh
      - -xc
      - sudo echo REGDOMAIN=US | sudo tee /etc/default/crda > /dev/null
    - - sh
      - -xc
      - sudo chown -R ubuntu:ubuntu /home/ubuntu
    write_files:
    - content: this is a ssh key
      path: /home/ubuntu/.ssh/id_rsa
      permissions: '0600'
    - content: this is a pub key
      path: /home/ubuntu/.ssh/id_rsa.pub
      permissions: '0644'

    """

    HEADER = "#cloud-config"

    def __init__(self, default=False):
        self.content = {}
        if default:
            self.__default__()

    def __str__(self):
        return Userdata.HEADER + '\n' + yaml.dump(self.content)

    def with_authorized_keys(self, keys=None):
        """
        Adds a list of authorized keys for default user ubuntu
        """
        if keys is None:
            raise Exception('keys arg supplied is None')
        if type(keys) != list:
            raise TypeError('Expected type of keys to be a list')

        if 'ssh_authorized_keys' not in self.content:
            self.content['ssh_authorized_keys'] = keys
        else:
            self.content['ssh_authorized_keys'] += keys
        return self

    def write(self, filename=None):
        """
        Writes a file to a location. Safe write for files on mounted partitions
        """
        if filename is None:
            raise Exception('filename arg supplied is None')
        tmp_location = path_expand('~/.cloudmesh/user-data.tmp')
        writefile(tmp_location, str(self))
        Console.info(f'Writing to {filename}')
        Shell.run(f'cat {tmp_location} | sudo tee {filename}')

    def with_ssh_password_login(self, ssh_pwauth=True):
        if ssh_pwauth is None:
            raise Exception('ssh_pwauth arg supplied is None')
        self.content['ssh_pwauth'] = ssh_pwauth
        return self

    def with_package_update(self, update=True):
        self.content['package_update'] = update
        return self

    def with_package_upgrade(self, upgrade=True):
        self.content['package_upgrade'] = upgrade
        return self

    def with_set_wifi_country(self, country=None):
        if country is None:
            raise Exception('the country arg supplied is none')

        cmd = f"sudo iw reg set {country}"
        self.with_bootcmd(cmd)

        cmd = f"sudo echo REGDOMAIN={country} | sudo tee /etc/default/crda > /dev/null"
        self.with_runcmd(cmd)
        return self

    def with_hosts(self, hosts=None):
        """
        Give a list of ip:hostname formatted strings, add to /etc/hosts
        """
        if not hosts:
            raise Exception('hosts arg passed is None')

        for hostpair in hosts:
            if ':' not in hostpair:
                raise Exception('Expected hosts to be a list of ip:hostname strings')
            ip, hostname = hostpair.split(':')
            self.with_bootcmd(cmd=f'echo {ip} {hostname} >> /etc/hosts')
        return self

    def with_bootcmd(self, cmd=None):
        if cmd is None:
            raise Exception('the command arg supplied is none')

        if 'bootcmd' in self.content:
            if type(cmd) == str:
                self.content['bootcmd'].append(cmd)
            elif type(cmd) == list:
                self.content['bootcmd'].extend(cmd)
        else:
            if type(cmd) == str:
                self.content['bootcmd'] = [cmd]
            elif type(cmd) == list:
                self.content['bootcmd'] = cmd
        return self

    def with_runcmd(self, cmd=None):
        if cmd is None:
            raise Exception('the command arg supplied is none')

        if 'runcmd' in self.content:
            if type(cmd) == str:
                self.content['runcmd'].append(cmd)
            elif type(cmd) == list:
                self.content['runcmd'].extend(cmd)
        else:
            if type(cmd) == str:
                self.content['runcmd'] = [cmd]
            elif type(cmd) == list:
                self.content['runcmd'] = cmd
        return self

    def with_access_point_bridge(self, priv_interface='eth0', ext_interface='wlan0'):
        """
        Uses iptables to configure an access point bridge where devices on priv_interface
        can route internet traffice through the priv_interface of the device towards ext_interface

        Often, this is practical for providing an entire cluster internet through a centralized point (the manager)
        """
        self.with_packages(packages=["iptables-persistent"])
        # Enable ipv4 forwarding and configure ip tables rules
        self.with_runcmd(cmd="sudo sysctl -w net.ipv4.ip_forward=1")\
            .with_runcmd(cmd="sudo sed -i 's/#net\\.ipv4\\.ip_forward=1/net.ipv4.ip_forward=1/' /etc/sysctl.conf")\
            .with_runcmd(cmd=f"sudo iptables -A FORWARD -i {priv_interface} -o {ext_interface} -j ACCEPT")\
            .with_runcmd(cmd=f"sudo iptables -A FORWARD -i {ext_interface} -o {priv_interface} "
                             "-m state --state ESTABLISHED,RELATED -j ACCEPT")\
            .with_runcmd(cmd=f"sudo iptables -t nat -A POSTROUTING -o {ext_interface} -j MASQUERADE")\
            .with_runcmd(cmd="iptables-save > /etc/iptables/rules.v4")  # nqa: E501

        return self

    def with_write_files(self,
                         encoding=None,
                         content=None,
                         owner=None,
                         path=None,
                         permissions=None):
        arguments = locals()
        arguments['self'] = None
        arguments = [(k, v) for k, v in arguments.items() if v is not None]

        if path is None:
            raise Exception('the path arg supplied is none')
        if content is None:
            raise Exception('the content supplied is none')

        # supports multiline strings, keeps newlines
        if '\n' in content:
            content = '|\n' + content

        file = {}
        for arg in arguments:
            file[arg[0]] = arg[1]

        if 'write_files' in self.content:
            self.content['write_files'].append(file)
        else:
            self.content['write_files'] = [file]

        return self

    def with_fix_user_dir_owner(self, user=None):
        if user is None:
            raise Exception('the user arg supplie is none')

        cmd = f'sudo chown -R {user}:{user} /home/{user}'
        self.with_runcmd(cmd=cmd)
        return self

    def with_packages(self, packages=None):
        """
        Given a list of packages or single package as string, add to the installation list. Can be called multiple times
        """
        if packages is None:
            raise Exception('ssh_pwauth arg supplied is None')
        if type(packages) != list and type(packages) != str:
            raise Exception('Type of packages expected to be a list or str')
        if 'packages' not in self.content:
            self.content['packages'] = []

        if type(packages) == list:
            self.content['packages'] += packages
        elif type(packages) == str:
            self.content['packages'] += [packages]
        return self

    def with_net_tools(self):
        """
        Adds net-tools and inetutils-traceroute to package installation list

        For useful network commands such as route and traceroute
        """
        self.with_packages(packages=['net-tools', 'inetutils-traceroute'])
        return self

    def with_locale(self, locale='en_US'):
        if locale is None:
            raise Exception('locale arg supplied is None')
        self.content['locale'] = locale
        return self

    def with_hostname(self, hostname=None):
        if hostname is None:
            raise Exception('hostname arg supplied is None')
        self.content['preserve_hostname'] = False
        self.content['hostname'] = hostname
        return self

    def with_default_user(self):
        """
        Include default user ubuntu with default password ubuntu. Prompt password change upon first access
        """
        if 'chpasswd' not in self.content:
            self.content['chpasswd'] = {}

        self.content['chpasswd']['expire'] = True
        self.content['chpasswd']['list'] = ['ubuntu:ubuntu']
        return self

    def __default__(self):
        """
        Set the default configuration the one that comes burnt with the ubuntu server OS

        Captured with
        $ grep -Fv \\# /{mountpoint}/user-data
        (Removes comments)
        """
        self.with_default_user().with_ssh_password_login()



from cloudmesh.common.Shell import Shell
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_pi
from cloudmesh.burn.util import os_is_windows


def get_ssid():
    ssid = ""
    try:
        if os_is_mac():
            command = "/System/Library/PrivateFrameworks/Apple80211.framework/Versions/Current/Resources/airport -I"
            r = Shell.run(command).replace("\t", "").splitlines()
            ssid = Shell.cm_grep(r, " SSID:")[0].split(":")[1].strip()
        elif os_is_linux():
            command = "iwgetid -r"
            ssid = Shell.run(command).strip()
        elif os_is_pi():
            command = "iwgetid -r"
            ssid = Shell.run(command).strip()
        elif os_is_windows():
            command = "netsh wlan show interfaces"
            r = Shell.run(command).replace("\t", "").splitlines()
            ssid = Shell.cm_grep(r, " SSID ")[0].split(":")[1].strip()
    except:  # noqa
        pass

    return ssid



import os
import time
from getpass import getpass

# from cloudmesh.common.debug import VERBOSE
from cloudmesh.burn.Imager import Imager
from cloudmesh.burn.burner.Burner import Burner
from cloudmesh.burn.burner.RaspberryBurner import Burner as RaspberryBurner
from cloudmesh.burn.burner.raspberryos import MultiBurner
from cloudmesh.burn.image import Image
from cloudmesh.burn.network import Network
from cloudmesh.burn.sdcard import SDCard
from cloudmesh.burn.ubuntu.configure import Configure
from cloudmesh.burn.util import os_is_linux
from cloudmesh.burn.util import os_is_mac
from cloudmesh.burn.util import os_is_pi
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.security import generate_strong_pass
from cloudmesh.common.util import Console
from cloudmesh.common.util import yn_choice
from cloudmesh.common.util import banner
from cloudmesh.inventory.inventory import Inventory
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command
from cloudmesh.shell.command import map_parameters
from cloudmesh.burn.usb import USB
from cloudmesh.common.debug import VERBOSE
from cloudmesh.burn.wifi.ssid import get_ssid
from cloudmesh.common.Host import Host
from cloudmesh.common.util import path_expand
from cloudmesh.common.Shell import Shell


class BurnCommand(PluginCommand):

    # noinspection PyBroadException
    @command
    def do_burn(self, args, arguments):
        """
        ::

            Usage:
              burn gui [--hostname=HOSTNAME]
                       [--ip=IP]
                       [--ssid=SSID]
                       [--wifipassword=PSK]
                       [--bs=BLOCKSIZE]
                       [--dryrun]
                       [--no_diagram]
              burn ubuntu NAMES [--inventory=INVENTORY] [--ssid=SSID] [-f]
              [--wifipassword=PSK] [-v] --device=DEVICE [--country=COUNTRY]
              [--upgrade]
              burn raspberry NAMES --device=DEVICE
                                  [--inventory=INVENTORY]
                                  [--ssid=SSID]
                                  [--wifipassword=PSK]
                                  [--country=COUNTRY]
                                  [--password=PASSWORD]
                                  [-v]
                                  [-f]
              burn firmware check
              burn firmware update
              burn install
              burn load --device=DEVICE
              burn format --device=DEVICE
              burn imager [TAG...]
              burn mount [--device=DEVICE] [--os=OS]
              burn unmount [--device=DEVICE] [--os=OS]
              burn network list [--ip=IP] [--used]
              burn network
              burn info [--device=DEVICE]
              burn image versions [--details] [--refresh] [--yaml]
              burn image ls
              burn image delete [--image=IMAGE]
              burn image get [--url=URL] [TAG...]
              burn backup [--device=DEVICE] [--to=DESTINATION]
              burn copy [--device=DEVICE] [--from=DESTINATION]
              burn shrink [--image=IMAGE]
              burn cluster --device=DEVICE --hostname=HOSTNAME
                           [--burning=BURNING]
                           [--ip=IP]
                           [--ssid=SSID]
                           [--wifipassword=PSK]
                           [--bs=BLOCKSIZE]
                           [--os=OS]
                           [-y]
                           [--imaged]
                           [--set_passwd]
              burn create [--image=IMAGE]
                          [--device=DEVICE]
                          [--burning=BURNING]
                          [--hostname=HOSTNAME]
                          [--ip=IP]
                          [--sshkey=KEY]
                          [--blocksize=BLOCKSIZE]
                          [--passwd=PASSWD]
                          [--ssid=SSID]
                          [--wifipassword=PSK]
                          [--format]
                          [--tag=TAG]
                          [--inventory=INVENTORY]
                          [--name=NAME]
                          [-y]
              burn sdcard [TAG...] [--device=DEVICE] [-y]
              burn set [--hostname=HOSTNAME]
                       [--ip=IP]
                       [--key=KEY]
                       [--keyboard=COUNTRY]
                       [--cmdline=CMDLINE]
              burn enable ssh
              burn wifi --ssid=SSID [--passwd=PASSWD] [--country=COUNTRY]
              burn check [--device=DEVICE]
              burn mac --hostname=HOSTNAME



            Options:
              -h --help              Show this screen.
              --version              Show version.
              --image=IMAGE          The image filename,
                                     e.g. 2019-09-26-raspbian-buster.img
              --device=DEVICE        The device, e.g. /dev/sdX
              --hostname=HOSTNAME    The hostnames of the cluster
              --ip=IP                The IP addresses of the cluster
              --key=KEY              The name of the SSH key file
              --blocksize=BLOCKSIZE  The blocksise to burn [default: 4M]
              --burning=BURNING      The hosts to be burned

            Arguments:
               TAG                   Keyword tags to identify an image

            Files:
              This is not fully thought through and needs to be documented
              ~/.cloudmesh/images
                Location where the images will be stored for reuse

            Description:
                cms burn create --inventory=INVENTORY --device=DEVICE --name=NAME

                    Will refer to a specified cloudmesh inventory file (see cms help inventory).
                    Will search the configurations for NAME inside of INVENTORY and will burn
                    to DEVICE. Supports parameter expansion.

                cms burn create --passwd=PASSWD

                     if the passwd flag is added the default password is
                     queried from the commandline and added to all SDCards

                     if the flag is omitted login via the password is
                     disabled and only login via the sshkey is allowed

              Network

                cms burn network list

                    Lists the ip addresses that are on the same network

                     +------------+---------------+----------+-----------+
                     | Name       | IP            | Status   | Latency   |
                     |------------+---------------+----------+-----------|
                     | Router     | 192.168.1.1   | up       | 0.0092s   |
                     | iPhone     | 192.168.1.4   | up       | 0.061s    |
                     | red01      | 192.168.1.46  | up       | 0.0077s   |
                     | laptop     | 192.168.1.78  | up       | 0.058s    |
                     | unkown     | 192.168.1.126 | up       | 0.14s     |
                     | red03      | 192.168.1.158 | up       | 0.0037s   |
                     | red02      | 192.168.1.199 | up       | 0.0046s   |
                     | red        | 192.168.1.249 | up       | 0.00021s  |
                     +------------+----------------+----------+-----------+

                cms burn network list [--used]

                    Lists the used ip addresses as a comma separated parameter
                    list

                       192.168.50.1,192.168.50.4,...

                cms burn network address

                    Lists the own network address

                     +---------+----------------+----------------+
                     | Label   | Local          | Broadcast      |
                     |---------+----------------+----------------|
                     | wlan0   | 192.168.1.12   | 192.168.1.255  |
                     +---------+----------------+----------------+

                cms burn firmware check

                    Checks if the firmware on the Pi is up to date

                cms burn firmware update

                    Checks and updates the firmware on the Pi

                cms burn install

                    Installs a program to shrink img files. THis is
                    useful, after you created a backup to make the
                    backup smaller and allow faster burning in case of
                    recovery

                    This command is not supported on MacOS

                cms burn load --device=DEVICE

                    Loads the sdcard into the USB drive. Thi sis similar to
                    loading a cdrom drive. It s the opposite to eject

                cms burn format --device=DEVICE

                    Formats the SDCard in the specified device. Be
                    careful it is the correct device.  cms burn info
                    will help you to identifying it

                cms burn mount [--device=DEVICE] [--os=OS]

                    Mounts the file systems available on the SDCard

                cms burn unmount [--device=DEVICE] [--os=OS]

                    Unmounts the mounted file systems from the SDCard

                cms burn info [--device=DEVICE]

                    Provides useful information about the SDCard

                cms burn image versions [--refresh] [--yaml]

                    The images that you like to burn onto your SDCard
                    can be cached locally with the image command.  The
                    available images for the PI can be found when
                    using the --refresh option. If you do not specify
                    it it reads a copy of the image list from our
                    cache

                cms burn image ls

                    Lists all downloaded images in our cache. You can
                    download them with the cms burn image get command

                cms burn image delete [--image=IMAGE]

                    Deletes the specified image. The name can be found
                    with the image ls command

                cms burn image get [--url=URL] [TAG...]

                    Downloads a specific image or the latest
                    image. The tag are a number of words separated by
                    a space that must occur in the tag that you find
                    in the versions command

                cms burn backup [--device=DEVICE] [--to=DESTINATION]

                    This command requires you to install pishrink previously with

                        cms burn install

                    Backs up a SDCard to the given location.

                cms burn copy [--device=DEVICE] [--from=DESTINATION]

                    Copies the file form the destination on the SDCard
                    this is the same as the SDCard command. we will in
                    future remove one

                cms burn shrink [--image=IMAGE]

                    Shrinks the size of a backup or image file that
                    is on your local file system. It can only be used
                    for .img files

                   This command is not supported on MacOS.

                cms burn create [--image=IMAGE]
                                [--device=DEVICE]
                                [--hostname=HOSTNAME]
                                [--ip=IP]
                                [--sshkey=KEY]
                                [--blocksize=BLOCKSIZE]
                                [--passwd=PASSWD]
                                [--ssid=SSID]
                                [--wifipassword=PSK]
                                [--format]

                    This command  not only can format the SDCard, but
                    also initializes it with specific values

                cms burn sdcard [TAG...] [--device=DEVICE]

                    this burns the sd card, see also copy and create

                cms burn set [--hostname=HOSTNAME]
                             [--ip=IP]
                             [--key=KEY]
                             [--mount=MOUNTPOINT]
                             [--keyboard=COUNTRY]
                             [--cmdline=CMDLINE]

                    Sets specific values on the sdcard after it
                    has ben created with the create, copy or sdcard
                    command

                    a --ssh is missing from this command

                cms burn enable ssh [--mount=MOUNTPOINT]

                    Enables the ssh server once it is booted

                cms burn wifi --ssid=SSID [--passwd=PASSWD] [--country=COUNTRY]

                    Sets the wifi ssid and password after the card
                    is created, copied, or the sdcard is used.

                    The option country option expects an ISO 3166-1
                    two digit country code. The default is "US" and
                    the option not required if suitable. See
                    https://en.wikipedia.org/wiki/ISO_3166-1 for other
                    countries.

                cms burn check [--device=DEVICE]

                    Lists the parameters that were set
                    with the set or create command

            Examples: ( \\ is not shown)

               > cms burn create --image=2019-09-26-raspbian-buster-lite
               >                 --device=/dev/mmcblk0
               >                 --hostname=red[5-7]
               >                 --ip=192.168.1.[5-7]
               >                 --sshkey=id_rsa

               > cms burn image get latest

               > cms burn image get https://downloads.raspberrypi.org/
               >   raspbian_lite/images/
               >   raspbian_lite-2018-10-11/2018-10-09-raspbian-stretch-lite.zip

               > cms burn image delete 2019-09-26-raspbian-buster-lite

        """
        map_parameters(arguments,
                       "details",
                       "refresh",
                       "device",
                       "dryrun",
                       "burning",
                       "hostname",
                       "ip",
                       "sshkey",
                       "blocksize",
                       "ssid",
                       "url",
                       "imaged",
                       "key",
                       "keyboard",
                       "passwd",
                       "wifipassword",
                       "version",
                       "to",
                       "os",
                       "country",
                       "inventory",
                       "name",
                       "bs",
                       "set_passwd",
                       "cmdline",
                       "upgrade",
                       "no_diagram")

        # arguments.MOUNTPOINT = arguments["--mount"]
        arguments.FORMAT = arguments["--format"]
        arguments.FROM = arguments["--from"]
        arguments.IMAGE = arguments["--image"]
        arguments.output = "table"  # hard code for now
        arguments.bs = arguments.bs or "4M"
        arguments.yes = arguments["-y"]
        if len(arguments.TAG) == 0:
            arguments.TAG = "latest"

        # VERBOSE(arguments)

        def execute(label, function):
            StopWatch.start(label)
            result = function
            StopWatch.stop(label)
            StopWatch.status(label, True)
            return result

        burner = Burner()
        sdcard = SDCard()

        if arguments.imager:

            arguments.TAG = arguments.TAG or ["latest-lite"]

            Console.msg(f"Tags: {arguments.TAG}")
            try:
                file = Imager.fetch(tag=arguments.TAG)
            except:  # noqa: E722
                pass

            try:
                Imager.launch(file=file)
            except Exception as e:
                Console.error(f"could not find image with the tag {arguments.TAG}\n\n{e}\n")

            return ""

        elif arguments.gui:

            from cloudmesh.burn.gui import Gui
            VERBOSE(arguments)
            g = Gui(hostname=arguments.hostname,
                    ip=arguments.ip,
                    dryrun=arguments.dryrun,
                    no_diagram=arguments.no_diagram)

            g.run()

            return ""

        elif arguments.raspberry:
            banner(txt="RaspberryOS Burn", figlet=True)

            names = Parameter.expand(arguments.NAMES)
            manager, workers = Host.get_hostnames(names)
            ssid = arguments['--ssid']
            wifipasswd = arguments['--wifipassword']

            if arguments.inventory:
                inv_path = path_expand(f'~/.cloudmesh/{arguments.inventory}.yaml')
                try:
                    burner = RaspberryBurner(inventory=inv_path)
                except:
                    Console.error('Burner Error')
                    return ""
            else:
                if workers:
                    worker_base_name = ''.join(
                        [i for i in workers[0] if not i.isdigit()])

                cluster_name = manager or worker_base_name
                inventory = path_expand(f'~/.cloudmesh/inventory-{cluster_name}.yaml')

                if not os.path.exists(inventory) or arguments['-f']:
                    if not manager:
                        Console.error("No inventory found. Can not create an "
                                      "inventory without a "
                                      "manager.")
                        return ""

                    _build_default_inventory(filename=inventory,
                                             manager=manager, workers=workers)

                try:
                    burner = RaspberryBurner(inventory=inventory)
                except:
                    Console.error('Burner Error')
                    return ""

                if manager:
                    if not ssid:
                        ssid = get_ssid()
                        if ssid == "":
                            Console.info('Could not determine SSID, skipping wifi '
                                         'config')
                            ssid = None
                    if not wifipasswd and ssid is not None:
                        wifipasswd = getpass(f"Using --SSID={ssid}, please "
                                             f"enter wifi password:")

            execute("burn raspberry", burner.multi_burn(
                names=arguments.NAMES,
                devices=arguments.device,
                verbose=arguments['-v'],
                password=arguments['--password'],
                ssid=ssid,
                wifipasswd=wifipasswd,
                country=arguments['--country']
            ))
            return ""

        elif arguments.ubuntu:
            banner(txt="Ubuntu Burn with cloud-init", figlet=True)
            names = Parameter.expand(arguments.NAMES)
            if len(Parameter.expand(arguments.device)) > 1:
                Console.error("Too many devices specified. Please only specify one")
                return ""

            if arguments.inventory:
                c = Configure(inventory=arguments.inventory, debug=arguments['-v'])
                inv = Inventory(filename=arguments.inventory)
            else:
                names = Parameter.expand(arguments.NAMES)
                manager, workers = Host.get_hostnames(names)
                if workers:
                    worker_base_name = ''.join(
                        [i for i in workers[0] if not i.isdigit()])

                cluster_name = manager or worker_base_name
                inventory = path_expand(f'~/.cloudmesh/inventory-{cluster_name}.yaml')

                if not os.path.exists(inventory) or arguments['-f']:
                    if not manager:
                        Console.error("No inventory found. Can not create an "
                                      "inventory without a "
                                      "manager.")
                        return ""

                    _build_default_inventory(filename=inventory,
                                             manager=manager,
                                             workers=workers,
                                             manager_image='ubuntu-20.10-64-bit',
                                             worker_image='ubuntu-20.10-64-bit')

                c = Configure(inventory=inventory, debug=arguments['-v'])
                inv = Inventory(filename=inventory)

                if manager:
                    if not arguments.ssid:
                        arguments.ssid = get_ssid()
                        if arguments.ssid == "":
                            Console.info('Could not determine SSID, skipping wifi '
                                         'config')
                            arguments.ssid = None
                    if not arguments.wifipassword and arguments.ssid is not None:
                        arguments.country = Shell.locale().upper()
                        arguments.wifipassword = getpass(f"Using --SSID="
                                                         f"{arguments.ssid} and "
                                                         f" --COUNTRY="
                                                         f"{arguments.country}, please "
                                                         f"enter wifi password:")

            # Probably not the best way to get the tag, but we assume all tags
            # are the same for each row for now
            if inv.has_host(names[0]):
                tag = inv.get(name=names[0], attribute='tag')
            else:
                Console.error(f'Could not find {names[0]} in inventory {inv.filename}')
                return ""
            if 'ubuntu' not in tag:
                Console.error("This command only supports burning ubuntu cards")
                return ""
            sdcard = SDCard(card_os="ubuntu")

            # Code below taken from arguments.sdcard
            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            # determine if we are burning a manager, as this needs to be done
            # first to get the ssh public key
            manager = False
            for name in names:
                if not inv.has_host(name):
                    Console.error(f'Could not find {name} in inventory {inv.filename}')
                    return ""
                service = inv.get(name=name, attribute='service')
                if service == 'manager' and not manager:
                    manager = name
                    # make manager first in names
                    names.remove(name)
                    names.insert(0, name)
                elif service == 'manager' and manager:
                    raise Exception('More than one manager detected in NAMES')

            for name in names:
                if not yn_choice(f'Is the card to be burned for {name} inserted?'):
                    if not yn_choice(f"Please insert the card to be burned for {name}. "
                                     "Type 'y' when done or 'n' to terminante"):
                        Console.error("Terminating: User Break")
                        return ""

                service = inv.get(name=name, attribute='service')
                # Make sure bridge is only enabled if WiFi enabled
                if service == 'manager':
                    services = inv.get(name=name, attribute='services')
                    if 'bridge' in services and not arguments.ssid:
                        Console.error('Service bridge can only be configured if WiFi'
                                      ' is enabled with --ssid and --wifipassword')
                        return ""
                    else:
                        enable_bridge = 'bridge' in services

                Console.info(f'Burning {name}')
                sdcard.format_device(device=arguments.device, yes=True)
                sdcard.unmount(device=arguments.device)
                sdcard.burn_sdcard(tag=tag, device=arguments.device, yes=True)
                sdcard.mount(device=arguments.device, card_os="ubuntu")
                if service == 'manager':
                    # Generate a private public key pair for the manager that will be persistently used
                    priv_key, pub_key = c.generate_ssh_key(name)
                    # Write priv_key and pub_key to /boot/id_rsa and /boot/id_rsa.pub
                    SDCard.writefile(filename=f'{sdcard.boot_volume}/id_rsa', content=priv_key)
                    SDCard.writefile(filename=f'{sdcard.boot_volume}/id_rsa.pub', content=pub_key)
                    c.build_user_data(name=name,
                                      country=arguments.country,
                                      upgrade=arguments.upgrade,
                                      with_bridge=enable_bridge).write(
                        filename=sdcard.boot_volume + '/user-data')
                    c.build_network_data(name=name, ssid=arguments.ssid, password=arguments.wifipassword)\
                        .write(filename=sdcard.boot_volume + '/network-config')
                else:
                    c.build_user_data(name=name, add_manager_key=manager,
                                      upgrade=arguments.upgrade).write(
                        filename=sdcard.boot_volume + '/user-data')
                    c.build_network_data(name=name).write(filename=sdcard.boot_volume + '/network-config')
                time.sleep(1)  # Sleep for 1 seconds to give ample time for writing to finish
                sdcard.unmount(device=arguments.device, card_os="ubuntu")

                Console.info("Remove card")

            Console.ok(f"Burned {len(names)} card(s)")
            return ""

        elif arguments.firmware and arguments.check:

            execute("firmware check", burner.firmware(action="check"))
            return ""

        elif arguments.firmware and arguments.update:

            execute("firmware update", burner.firmware(action="update"))
            return ""

        if arguments.check:

            execute("check", burner.check(device=arguments.device))
            return ""

        elif arguments.versions and arguments['image']:

            StopWatch.start("image versions")

            result = Image.create_version_cache(refresh=arguments["--refresh"])

            output = "table"
            if arguments["--yaml"]:
                output = "yaml"

            order = ["tag", 'date', "os", "type", 'version']
            header = ["Tag", 'Date', "OS", "Type", 'Version']
            if arguments.details:
                order = ["tag", 'date', "os", "type", 'version', "url"]
                header = ["Tag", 'Date', "OS", "Type", 'Version', "Url"]

            print(Printer.write(result, order=order, header=header, output=output))

            StopWatch.stop("image versions")
            StopWatch.status("image versions", True)
            return ""

        elif arguments.load:
            execute("load", sdcard.load_device(device=arguments.device))
            return ""

        elif arguments["format"]:  # as format is a python word, we need to use an index

            execute("format", sdcard.format_device(device=arguments.device, unmount=True))
            return ""

        elif arguments.network and arguments["list"]:

            if os_is_mac():
                Console.error("Not yet implemented on MacOS")
                return ""

            ip = arguments.ip or Network.address()[0]['local']

            details = Network.nmap(ip=ip)

            if arguments.used:

                print(','.join([x['ip'] for x in details]))

            else:
                print(Printer.write(
                    details,
                    order=['name', "ip", "status", "latency", ],
                    header=['Name', "IP", "Status", "Latency", ]
                )
                )
            return ""

        elif arguments.network:

            if os_is_mac():
                Console.error("Not yet implemented on MacOS")
                return ""

            # print (Network.nmap())
            details = Network.address()

            print(Printer.write(
                details,
                order=['label', "local", "broadcast"],
                header=["Label", "Local", "Broadcast"]
            )
            )
            return ""

        elif arguments.wifi:

            password = arguments.passwd
            ssid = arguments.ssid or get_ssid()
            country = arguments.country

            if password is None:
                password = getpass("Please enter the Wifi password or enter "
                                   "for no password: ")

            if os_is_mac():
                host = "macos"
            elif os_is_linux():
                host = "linux"
            elif os_is_pi():
                host = "raspberry"
            else:
                Console.error("This command is not yet implemented for your OS")
                return ""

            burner.configure_wifi(ssid, psk=password, country=country, host=host)

            return ""

        elif arguments.info:

            output = arguments.output or "table"
            card = SDCard()
            execute("info", card.info(output=output))
            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            return ""

        elif arguments.install:

            if os_is_mac():
                Console.error("Not yet implemented on MacOS")
                return ""

            execute("install", burner.install())
            return ""

        elif arguments.shrink:

            if os_is_mac():
                Console.error("Not yet implemented on MacOS")
                return ""

            execute("shrink", burner.shrink(image=arguments.IMAGE))
            return ""

        elif arguments.backup:
            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""
            execute("backup", sdcard.backup(device=arguments.device, to_file=arguments.to))
            return ""

        elif arguments["copy"]:  # as copy is a reserved word we need to use the index
            USB.check_for_readers()
            execute("copy", sdcard.copy(device=arguments.device, from_file=arguments.FROM))
            return ""

        elif arguments.sdcard:

            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            if arguments.device is None:
                card = SDCard()
                card.info()
                Console.error("Please specify a device")
                return ""

            arguments.TAG = arguments.TAG or ["latest-lite"]
            if any("ubuntu" in tag for tag in arguments.TAG):
                sdcard = SDCard(card_os="ubuntu")

            execute("format", sdcard.format_device(device=arguments.device, unmount=True))
            execute("unmount", sdcard.unmount(device=arguments.device))

            execute("sdcard", sdcard.burn_sdcard(tag=arguments.TAG, device=arguments.device, yes=arguments.yes))
            return ""

        elif arguments.mount:

            if arguments.device is None:
                card = SDCard
                card.info()
                Console.error("Please specify a device")
                return ""

            execute("mount", sdcard.mount(device=arguments.device, card_os=arguments.os))
            return ""

        elif arguments.unmount:

            card = SDCard(card_os=arguments.os)
            execute("unmount", card.unmount(device=arguments.device, card_os=arguments.os))
            return ""

        elif arguments.mac:

            hostnames = Parameter.expand(arguments.hostname)

            execute("mac", burner.mac(hostnames=hostnames))
            return ""

        elif arguments.set:

            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            if arguments.hostname:
                execute("set hostname", burner.set_hostname(arguments.hostname))

            if arguments.ip:
                execute("set ip", burner.set_static_ip(arguments.ip))

            if arguments.key:
                execute("set key", burner.set_key(arguments.key))

            if arguments.keyboard:
                execute("set keyboard", burner.keyboard(country=arguments.keyboard))

            if arguments.cmdline:
                execute("set cmdline", burner.set_cmdline(arguments.cmdline))

            return ""

        elif arguments.enable and arguments.ssh:
            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            execute("enable ssh", burner.enable_ssh())
            return ""

        # elif arguments.versions and arguments.image:
        #    image = Image()

        elif arguments.ls and arguments['image']:
            execute("image ls", Image().ls())
            return ""

        elif arguments.delete and arguments.IMAGE:
            execute("image rm", Image().rm(arguments.IMAGE))
            return ""

        elif arguments["get"] and arguments['image'] and arguments["--url"]:
            image = Image()
            execute("image fetch", image.fetch(url=arguments.url))
            return ""

        elif arguments["get"] and arguments['image'] and arguments["TAG"]:

            tag = arguments["TAG"]
            if "latest" in tag and ("full" in tag or "lite" in tag):
                result = Image.create_version_cache(refresh=arguments["--refresh"])

            image = Image()
            execute("image fetch", image.fetch(tag=arguments["TAG"]))
            return ""

        elif arguments["get"] and arguments['image']:
            image = Image()
            execute("image fetch", image.fetch(tag="latest"))
            return ""

        elif arguments.cluster:

            # is true when
            #
            # cms burn cluster --hostname=red,red00[1-2]
            #                  --device=/dev/sdb
            #                  --ip=10.1.1.[1-3]
            #                  --ssid=myssid
            #                  --wifipassword=mypass
            #

            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""
            execute("cluster", burner.cluster(arguments=arguments))
            return ""

        elif arguments.create and arguments.inventory:
            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            if not os_is_pi():
                print()
                Console.error("This command has only been written for a  Raspberry Pis. "
                              "Terminating for caution")
                print()
                if yn_choice("Continue anyways?"):
                    pass
                else:
                    return
            if not arguments.name:
                Console.error("Missing --name parameter. See cms help burn for usage")
                return ""
            if not arguments.device:
                Console.error("Missing --device parameter. See cms help burn for usage")
                return ""

            StopWatch.start("burn inventory")
            multi_burner = MultiBurner()
            # Perhaps we want to change the path at some point
            inventory = f"~/.cloudmesh/{arguments.inventory}"
            multi_burner.burn_inventory(
                inventory=inventory,
                name=arguments.name,
                device=arguments.device,
                yes=arguments.yes,
                passwd=arguments.passwd
            )
            StopWatch.stop("burn inventory")
            StopWatch.status("burn inventory", True)
            StopWatch.benchmark(sysinfo=False, csv=False)
            return ""

        elif arguments.create:

            try:
                USB.check_for_readers()
            except Exception as e:
                print()
                Console.error(e)
                print()
                return ""

            if arguments["--passwd"]:
                passwd = arguments["--passwd"]
            elif "PASSWD" in os.environ:
                passwd = os.environ["PASSWD"]
            else:
                passwd = generate_strong_pass()

            psk = None
            if arguments["--ssid"]:
                ssid = arguments["--ssid"]
                if arguments["--wifipassword"]:
                    psk = arguments["--wifipassword"]
                else:
                    psk = None
            else:
                if arguments["--wifipassword"]:
                    print("Can't have wifi password with no ssid")
                    return
                else:
                    ssid = None

            image = 'latest' or arguments.IMAGE

            dev = os.environ['DEV'] if 'DEV' in os.environ else None
            devices = arguments["--device"] or dev or None

            if devices is not None:
                devices = Parameter.expand_string(devices)

            hostnames = Parameter.expand(arguments.hostname)

            if arguments.burnimg is None:
                burning = hostnames
            else:
                burning = arguments.burning

            VERBOSE(arguments)

            ips = None if not arguments.ip else Parameter.expand(arguments.ip)
            key = arguments.sshkey
            tag = arguments['--tag']

            if os_is_pi() or os_is_linux():
                blocksize = arguments.blocksize
                StopWatch.start("total")

                multi = MultiBurner()
                multi.burn_all(
                    burning=burning,
                    image=image,
                    device=devices,
                    blocksize=blocksize,
                    progress=True,
                    hostnames=hostnames,
                    # not difference between names and name, maybe we should align
                    ips=ips,
                    key=key,
                    password=passwd,
                    ssid=ssid,
                    psk=psk,
                    tag=tag,
                    yes=arguments.yes
                )

                StopWatch.stop("total")
                StopWatch.status("total", True)

                StopWatch.benchmark(sysinfo=False, csv=False)
            else:
                Console.error("This command is only supported ona Pi and Linux")
            return ""

        Console.error("see manual page: cms help burn")
        return ""


def _build_default_inventory(filename, manager, workers, ips=None,
                             manager_image='latest-lite',
                             worker_image='latest-lite'):
    # cms inventory add red --service=manager --ip=10.1.1.1 --tag=latest-lite
    # --timezone="America/Indiana/Indianapolis" --locale="us"
    # cms inventory set red services to "bridge" --listvalue
    # cms inventory add "red0[1-3]" --service=worker --ip="10.1.1.[2-4]"
    # --router=10.1.1.1 --tag=latest-lite  --timezone="America/Indiana/Indianapolis" --locale="us"
    # cms inventory set "red0[1-3]" dns to "8.8.8.8,8.8.4.4" --listvalue

    Console.info("No inventory found or forced rebuild. Buidling inventory "
                 "with defaults.")
    Shell.execute("rm", arguments=[
                  '-f', filename])
    i = Inventory(filename=filename)
    timezone = Shell.timezone()
    locale = Shell.locale()
    manager_ip = ips[0] if ips else '10.1.1.1'
    image = manager_image
    element = {}
    element['host'] = manager
    element['status'] = 'inactive'
    element['service'] = 'manager'
    element['ip'] = manager_ip
    element['tag'] = image
    element['timezone'] = timezone
    element['locale'] = locale
    element['services'] = ['bridge', 'wifi']
    element['keyfile'] = '~/.ssh/id_rsa.pub'
    i.add(**element)
    i.save()

    last_octet = 2
    index = 1
    if workers is not None:
        for worker in workers:
            ip = ips[index] if ips else f'10.1.1.{last_octet}'
            image = worker_image
            element = {}
            element['host'] = worker
            element['status'] = 'inactive'
            element['service'] = 'worker'
            element['ip'] = ip
            element['tag'] = image
            element['timezone'] = timezone
            element['locale'] = locale
            element['router'] = manager_ip
            element['dns'] = ['8.8.8.8', '8.8.4.4']
            element['keyfile'] = '~/.ssh/id_rsa.pub'
            i.add(**element)
            i.save()
            last_octet += 1
            index += 1

    i.save()
    print(i.list(format="table"))



import json
import socket
import subprocess


# TODO: there seems to be some overlap with hardware.py and some functions in
#       hardware introduced that are not used here such as getfqdn and hostname.
#       Should it be removed form hardware, or should this be changed here to
#       use the methods from hardware?
#
# TODO: get method is not implemented

class Network:

    # noinspection PyBroadException,PyBroadException
    @staticmethod
    def address():
        hostname = socket.gethostname()
        full = socket.getfqdn()

        # ipv4
        result = []
        try:
            details = json.loads(subprocess.getoutput("ip -json a"))
        except Exception as e:  # noqa: F841
            details = None
        for entry in details:
            try:
                addresses = entry['addr_info']
                for n in addresses:
                    if entry['ifname'] not in ['lo']:
                        element = {
                            'ip': socket.gethostbyname(hostname),
                            'hostname': hostname,
                            'ifname': entry['ifname'],
                            'fullname': full,
                            'ipbyname': socket.gethostbyname(hostname),

                        }
                        # print ("nnn", n)
                        element.update(n)
                        if 'broadcast' in n.keys():
                            result.append(element)
            except Exception as e:  # noqa: F841
                pass
        return result

    @staticmethod
    def nmap(ip=None):
        if ip is None:
            ip = Network.address()['ip']
        mask = ip.rsplit(".", 1)[0]
        command = f"nmap -sP {mask}.*"
        print(command)
        result = subprocess.getoutput(command)
        result = result.replace("Nmap scan report for ", "")
        result = result.replace("Host is ", "")
        result = result.replace(" latency.", "")

        result = result.splitlines()[1:]
        details = []
        for i in range(0, int(len(result) / 2)):
            position = 2 * i
            a = result[position]
            b = result[position + 1]
            line = f"{a} {b}"
            if "(" not in a:
                line = f"unkown {a} {b}"

            line = line.replace("(", "")
            line = line.replace(")", "")

            attributes = line.split()
            details.append({
                'name': attributes[0],
                'ip': attributes[1],
                'status': attributes[2],
                'latency': attributes[3],
                'line': line
            })

        return details



"""
Implementation of a function the set the WIFI configuration.
This function is primarily developed for a Raspberry PI
"""

from cloudmesh.burn.wifi.raspberryos import Wifi as WifiRaspberryOs
from cloudmesh.burn.wifi.ubuntu import Wifi as WifiUbuntu


# noinspection PyPep8Naming
def Wifi(card_os="raspberry"):
    if card_os == "raspberry":
        return WifiRaspberryOs
    else:
        return WifiUbuntu



import os
import socket


class Hardware(object):

    @staticmethod
    def is_pi():
        """
        Checks if its called on a PI

        :return: returns true if this is called on a Pi
        :rtype: bool
        """
        return os.uname()[4][:3] == 'arm' and 'Raspberry' in Hardware.model()

    @staticmethod
    def get_mac(interface='eth0'):
        """
        Get the mac address

        :param interface: the network interface name
        :type interface: str
        :return: mac address
        :rtype: str
        """
        # noinspection PyBroadException
        try:
            address = open(f'/sys/class/net/{interface}/address').read()
        except Exception as e:  # noqa: F841
            address = "00:00:00:00:00:00"
        return address[0:17]

    @staticmethod
    def get_ethernet():
        """
        TODO: describe

        :return:
        :rtype:
        """
        interface = None
        # noinspection PyBroadException
        try:
            for root, dirs, files in os.walk('/sys/class/net'):
                for directory in dirs:
                    if directory[:3] == 'enx' or directory[:3] == 'eth':
                        interface = directory
        except Exception as e:  # noqa: F841
            interface = "None"
        return interface

    @staticmethod
    def model():
        """
        TODO: describe

        :return:
        :rtype:
        """
        # noinspection PyBroadException
        try:
            model_str = open('/sys/firmware/devicetree/base/model').read()
        except Exception as e:  # noqa: F841
            model_str = "unkown"

        return model_str

    @staticmethod
    def hostname():
        """
        The hostname

        :return: the hostname
        :rtype: str
        """
        return socket.gethostname()

    @staticmethod
    def fqdn():
        """
        TODO: describe

        :return:
        :rtype:
        """
        return socket.getfqdn()


"""
eth0 = Hardware.get_mac('eth0')
wan = Hardware.get_mac('wlan0')

print(eth0)

print(wan)

print(os.uname()[4][:3] == 'arm')

print(Hardware.model())
print(Hardware.is_pi())
print(Hardware.hostname())
print(Hardware.fqdn())
"""



import glob
import os
import subprocess

import humanize
import requests

import usb as usb_device
from cloudmesh.burn.util import os_is_mac
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.sudo import Sudo
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.util import writefile


# from cloudmesh.burn.util import os_is_linux
# from cloudmesh.burn.util import os_is_pi


def _get_attribute(attribute, lines):
    for line in lines:
        if attribute in line:
            return line.split(attribute, 1)[1].strip()
    return None


class USB(object):
    """
    Class to make interacting with USB devices such as SDCard writers easier.
    """

    def __init__(self):
        self.vendors = None

    # noinspection PyBroadException
    def get_product(self, vendor=None, product=None):
        """
        internal method used to retrieve the vendor, product string
        :param vendor: the vendor name
        :type vendor: str
        :param product: the product name
        :type product: str
        :return: the vendor product string
        :rtype: str
        """
        try:
            return self.vendors[vendor][product]
        except Exception as e:  # noqa: F841
            return "unkown"

    # noinspection PyBroadException
    def load_vendor_description(self):
        """
        Creates a dict from the usb devices that are detected.
        Attributes for each entry include

            'vendor_id'
            'product_id'
            'vendor'
            'product'

        :return: a dict of vendor specifications for the USB devices
        :rtype: dict
        """
        self.get_vendor()
        data = {}
        product_id = None
        vendor_id = None
        vendor = None
        content = self.get_vendor().splitlines()
        for line in content:
            try:
                other_devices = ['C', 'AT', 'HID', 'R', 'BIAS', 'PHY', 'HUT',
                                 'L', 'HCC', 'VT']
                other_dev = False
                for dev in other_devices:
                    if line.startswith(dev):
                        other_dev = True

                first_word = line.strip().split()[0]

                if line.startswith("#") or line.startswith("]") or line is None or other_dev:
                    continue
                elif not line.startswith("\t") and len(first_word) == 4:
                    vendor_id, vendor = line.strip().split(" ", 1)
                    data[vendor_id] = {}
                elif line.startswith("\t") and len(first_word) == 4:
                    product_id, product = line.strip().split(" ", 1)
                    data[vendor_id][product_id] = {
                        'vendor_id': vendor_id,
                        'product_id': product_id,
                        'vendor': vendor,
                        'product': product
                    }

            except:  # noqa: E722
                pass
        self.vendors = data
        return data

    def get_vendor(self):
        """
        Retrieves the names of vendors from linux-usb.org

        :return: the content of the file
        :rtype: str
        """
        filename = 'usb.ids'
        full_path = path_expand(f"~/.cloudmesh/cmburn/{filename}")
        if not os.path.isfile(full_path):
            r = requests.get(f'http://www.linux-usb.org/{filename}')
            content = r.text
            writefile(full_path, content)
        else:
            content = readfile(full_path)
        return content

    @staticmethod
    def get_devices():
        """
        finds all devices starting with /dev/sd?

        :return: list
        :rtype: list of devices
        """
        if os_is_mac():
            return glob.glob("/dev/disk?")
        else:
            return glob.glob("/dev/sd?")

    @staticmethod
    def fdisk(dev):
        """
        calls fdisk on the specified device

        :param dev: device, example /dev/sdz
        :type dev: str
        :return: the output from the fdisk command
        :rtype: str
        """
        if os_is_mac():
            raise NotImplementedError("fdisk -l not supported on MacOS")
        else:
            Sudo.password()
            return subprocess.getoutput(f"sudo fdisk -l {dev}")

    # noinspection PyBroadException,PyBroadException
    @staticmethod
    def get_from_usb():
        """
        TODO: explain difference to get_from_lsusb

        Finds the information about attached USB devices from lsusb
        Attributes of the devices found include

            'comment'
            "hVendor"
            "hProduct"
            "search"

        :return: list of dicts
        :rtype: list
        """
        try:
            v = USB()
            v.load_vendor_description()
        except:  # noqa: E722
            pass

        def h(d, a):
            v = hex(d[a])
            return v.replace("0x", "0")

        lsusb = USB.get_from_lsusb()
        if len(lsusb) == 0:
            Console.error("We could not find your USB reader in the list of known readers")
            return None
        busses = usb_device.busses()

        details = []
        for bus in busses:
            devices = bus.devices
            for dev in devices:
                data = dev.__dict__
                data.update(dev.dev.__dict__)
                data['comment'] = lsusb[f"{dev.bus}-{dev.address}"]["comment"]
                del data['configurations']
                try:
                    vendor = f'{data["idVendor"]:04x}'
                    product = f'{data["idProduct"]:04x}'
                    vendor_str = v.vendors[vendor][product]['vendor']
                    device_str = v.vendors[vendor][product]['product']
                    data["hVendor"] = vendor_str
                    data["hProduct"] = device_str
                except:  # noqa: E722
                    data["hVendor"] = h(data, "idVendor")
                    data["hProduct"] = h(data, "idProduct")
                data["search"] = "tbd"
                details.append(data)
        return details

    # noinspection PyBroadException
    @staticmethod
    def get_from_lsusb():
        """
        Gets information about USB devices from lsusb. Attributes include

            'bus'
            'addr'
            'ivendor'
            'iproduct'
            'vendor'
            'product'
            'comment'

        :return: list of dicts
        :rtype: list of dicts
        """
        lsusb = subprocess.getoutput("lsusb").splitlines()

        found = {}
        for line in lsusb:
            line = line.replace("Bus ", "")
            line = line.replace("Device ", "")
            line = line.replace("ID ", "")
            line = line.replace(":", "", 1)
            line = line.replace(":", " ", 1)
            content = line.split(" ")

            try:
                bus = int(content[0])
                addr = int(content[1])

                data = {
                    'bus': bus,
                    'addr': addr,
                    'ivendor': int(content[2], 16),
                    'iproduct': int(content[3], 16),
                    'vendor': content[2],
                    'product': content[3],
                    'comment': ' '.join(content[4:])
                }
                found[f"{bus}-{addr}"] = data
            except Exception as e:  # noqa: F841
                pass

        return found

    # noinspection PyBroadException
    @staticmethod
    def get_from_dmesg(pluggedin=True):
        """
        Get information for USB and other direct attached devices from
        dmsg. It includes information such if the device is readable or writable,
        and if a card is formatted with FAT32.

        :param pluggedin: Only listed the plugged in USB devices
        :type pluggedin: bool
        :return: list of dicts
        :rtype: list of dicts
        """
        try:
            lines = subprocess.getoutput("dmesg -t").splitlines()
        except:
            return None

        details = {}
        for line in lines:
            if line.startswith("scsi") and "Direct-Access" in line:
                sdci, key, what = line.split(" ", 2)
                comment = line.split("Direct-Access")[1].strip()
                what = " ".join(comment.split("  ")[:2])
                if key not in details:
                    details[key] = {}
                details[key]["key"] = key
                details[key]["direct-access"] = True
                details[key]["info"] = what
                device, a, b, bus, rest = key.split(":")
                details[key]["device"] = device
                details[key]["bus"] = bus

            elif line.startswith("sd") and " sg" in line:
                sg = line.split(" sg")[1].split(" ", 1)[0]
                details[key]["sg"] = sg

            elif line.startswith("sd") and "] " in line:
                prefix, key, device, comment = line.split(" ", 3)
                if key not in details:
                    details[key] = {}
                details[key]["key"] = key
                if "Attached SCSI removable disk" in comment:
                    try:
                        details[key]["removable"] = True
                    except:  # noqa: E722
                        details[key]["removable"] = False
                if "logical blocks:" in comment:
                    size = comment.split("blocks:")[1]
                    details[key]["size"] = size.strip().replace("(",
                                                                "").replace(")",
                                                                            "")
                if "Write Protect is" in comment:
                    details[key]["writeable"] = "off" in comment
                else:
                    details[key]["writeable"] = True
                name = details[key]["name"] = device.replace("[", "").replace(
                    "]", "")
                # TODO:
                # This line was commented out previously, causing the program to fail. Why was this done?
                details[key]["dev"] = f"/dev/{name}"
                _fdisk = USB.fdisk(name)
                details[key]['readable'] = "cannot open" in _fdisk
                details[key]['empty'] = "linux" in _fdisk
                details[key]['formatted'] = "FAT32" not in _fdisk
                details[key]['active'] = os.path.exists(details[key]['dev'])
        # remove opbets without size

        found = []
        for name in details:
            entry = details[name]
            if 'size' in list(entry.keys()) and (entry['active'] or not pluggedin):
                found.append(entry)

        return found

    @staticmethod
    def check_for_readers():
        if os_is_mac():
            readers = USB.get_dev_from_diskutil()
            if len(readers) == 0:
                print()
                Console.error("Please make sure the reader and the SDCard "
                              "are properly inserted")
                raise ValueError("No card found")
            elif len(readers) > 1:
                print()
                Console.error("At this time we only support one SDCard "
                              "reader/writer for MacOS")
                raise ValueError("Too many cards found")

    @staticmethod
    def get_dev_from_diskutil():
        if os_is_mac():
            import plistlib
            external = subprocess.check_output("diskutil list -plist external".split(" "))

            r = dict(plistlib.loads(external))

            details = []

            if len(r['AllDisksAndPartitions']) == 0:
                Console.error("No partition found")
                return ""

            else:
                for dev in r['AllDisksAndPartitions']:
                    details.append(dev['DeviceIdentifier'])
                return details
        else:
            return None

    @staticmethod
    def get_from_diskutil(device=None):
        import plistlib
        external = subprocess.check_output("diskutil list -plist external".split(" "))

        r = dict(plistlib.loads(external))

        details = []

        if len(r['AllDisksAndPartitions']) == 0:
            Console.error("No partition found")
            return ""

        no = 0
        for cards in r['AllDisksAndPartitions']:

            try:
                for partition in r['AllDisksAndPartitions'][no]['Partitions']:
                    if 'MountPoint' not in partition:
                        partition['MountPoint'] = None
                    if partition['Content'] == 'Linux':
                        partition['Content'] = 'ext4'
                    elif partition['Content'] == 'Windows_FAT_32':
                        partition['Content'] = 'FAT32'

                    info = partition['MountPoint']
                    entry = {
                        "dev": f"/dev/{partition['DeviceIdentifier']}",
                        "active": info is not None,
                        "info": info,
                        "readable": info is not None,
                        "formatted": partition['Content'],
                        "empty": partition['Size'] == 0,
                        "size": humanize.naturalsize(partition['Size']),
                        "direct-access": True,
                        "removable": True,
                        "writeable": 'VolumeName' in partition
                    }
                    if device is None or device in entry["dev"]:
                        details.append(entry)
                no = no + 1
            except KeyError as e:  # noqa: F841
                Console.warning("No partitions found for device")
                partition = r['AllDisksAndPartitions'][no]
                entry = {
                    "dev": f"/dev/{partition['DeviceIdentifier']}",
                    "active": False,
                    "info": "Not Formatted",
                    "readable": False,
                    "formatted": False,
                    "empty": partition['Size'] == 0,
                    "size": partition['Size'],
                    "direct-access": True,
                    "removable": True,
                    "writeable": 'VolumeName' in partition
                }
                details.append(entry)

        return details

    @staticmethod
    def print_details(details, order=None, header=None, output="table"):
        if order is None:
            order = ["dev",
                     "info",
                     "formatted",
                     "size",
                     "active",
                     "readable",
                     "empty",
                     "direct-access",
                     "removable",
                     "writeable"
                     ],
        if header is None:
            header = ["Path",
                      "Info",
                      "Formatted",
                      "Size",
                      "Plugged-in",
                      "Readable",
                      "Empty",
                      "Access",
                      "Removable",
                      "Writeable"
                      ],
        # print(Printer.write(details, order=order, header=header, output=output))
        print(Printer.write(details, order=order[0], header=header[0]))
